"""
University of West Bohemia Air Traffic Control Communication (UWB-ATCC)

Šmídl, Luboš, 2011, Air Traffic Control Communication, LINDAT/CLARIAH-CZ digital library at the Institute of Formal and Applied Linguistics (ÚFAL), Faculty of Mathematics and Physics, Charles University, http://hdl.handle.net/11858/00-097C-0000-0001-CCA1-0.

Corpus contains recordings of communication between air traffic controllers and pilots. The speech is manually transcribed and labeled with the information about the speaker (pilot/controller, not the full identity of the person). The corpus is currently small (20 hours) but we plan to search for additional data next year. The audio data format is: 8kHz, 16bit PCM, mono.
"""

import hashlib
import logging
import re
import shutil
import unicodedata
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import (
    Pathlike,
    is_module_available,
    resumable_download,
    safe_extract_rar,
)


def download_uwb_atcc(
    target_dir: Pathlike = ".", force_download: Optional[bool] = False
) -> Path:
    if not is_module_available("rarfile"):
        raise ImportError("Please 'pip install rarfile' first.")
    import rarfile

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    dataset_name = "ZCU_CZ_ATC"
    rar_path = target_dir / f"{dataset_name}.rar"
    corpus_dir = target_dir / dataset_name
    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping {dataset_name} because {completed_detector} exists.")
        return corpus_dir
    resumable_download(
        f"https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11858/00-097C-0000-0001-CCA1-0/{dataset_name}.rar",
        filename=rar_path,
        completed_file_size=584245376,
        force_download=force_download,
    )
    if (
        hashlib.md5(open(rar_path, "rb").read()).hexdigest()
        != "44b4ea6ffe0ac0bf8fd29f14a735d23a"
    ):
        raise RuntimeError("MD5 checksum does not match")
    shutil.rmtree(corpus_dir, ignore_errors=True)
    with rarfile.RarFile(rar_path) as rar:
        safe_extract_rar(rar, path=corpus_dir)
    completed_detector.touch()
    return corpus_dir


def strip_accents(s):
    return "".join(
        c for c in unicodedata.normalize("NFD", s) if unicodedata.category(c) != "Mn"
    )


ABBREVIATIONS = {
    ".": "DECIMAL",
    "..": "DECIMAL DECIMAL",
    "FL": "FLIGHT LEVEL",
    "Fl": "FLIGHT LEVEL",
    "LLZ": "LOCALIZER",
    "PR": "PAPA ROMEO",
    "RWY": "RUNWAY",
}

UNKNOWN_ABBREVIATIONS = (
    "HDO",
    "VOZ",
    "VLM",
)

PHONETIC_ALPHABET = {
    "A": "ALFA",
    "B": "BRAVO",
    "C": "CHARLIE",
    "D": "DELTA",
    "E": "ECHO",
    "F": "FOXTROT",
    "G": "GOLF",
    "H": "HOTEL",
    "I": "INDIA",
    "J": "JULIETT",
    "K": "KILO",
    "L": "LIMA",
    "M": "MIKE",
    "N": "NOVEMBER",
    "O": "OSCAR",
    "P": "PAPA",
    "Q": "QUEBEC",
    "R": "ROMEO",
    "S": "SIERRA",
    "T": "TANGO",
    "U": "UNIFORM",
    "V": "VICTOR",
    "W": "WHISKEY",
    "X": "XRAY",
    "Y": "YANKEE",
    "Z": "ZULU",
}

INDIVIDUALLY_PRONOUNCED = (
    "ATR",
    "CRJ",
    "CSA",
    "CTO",
    "DEM",
    "DME",
    "EFC",
    "IFR",
    "ILS",
    "KLM",
    "QNH",
    "TMA",
    "UPS",
    "VFR",
    "VMC",
    "VOR",
)

FIX_TYPOS = {
    "ACCELARATING": "ACCELERATING",
    "ACCPET": "ACCEPT",
    "ACTUALY": "ACTUALLY",
    "AFETRNOON": "AFTERNOON",
    "AFFRIM": "AFFIRM",
    "AFTENOON": "AFTERNOON",
    "AFTERNON": "AFTERNOON",
    "AIRBORN": "AIRBORNE",
    "ALLRIGHT": "ALL RIGHT",
    "ALTITUED": "ALTITUDE",
    "APPORACH": "APPROACH",
    "APPORACHING": "APPROACHING",
    "APPRAOCH": "APPROACH",
    "APPROCHING": "APPROACHING",
    "APPRON": "APRON",
    "APROVED": "APPROVED",
    "APROXIMATELY": "APPROXIMATELY",
    "APROXIMETLY": "APPROXIMATELY",
    "AUSRTIAN": "AUSTRIAN",
    "AUSTRAIN": "AUSTRIAN",
    "AVAILBALE": "AVAILABLE",
    "AVALIABLE": "AVAILABLE",
    "AVIALABLE": "AVAILABLE",
    "BOARDLINE": "BROAD LINE",
    "BRUSSELES": "BRUSSELS",
    "BRUSSELS": "BRUSSELS",
    "CANCELED": "CANCELLED",
    "CANCELING": "CANCELLING",
    "CHALENGER": "CHALLENGER",
    "CHECH": "CZECH",
    "CIMB": "CLIMB",
    "CIMBING": "CLIMBING",
    "CLEARD": "CLEARED",
    "CLEARENCE": "CLEARANCE",
    "CLIBM": "CLIMB",
    "CLIMBIN": "CLIMBING",
    "CLMBING": "CLIMBING",
    "COMMING": "COMING",
    "CONACT": "CONTACT",
    "CONATACT": "CONTACT",
    "CONNTINUE": "CONTINUE",
    "CONTAC": "CONTACT",
    "CONTACE": "CONTACT",
    "CONTATC": "CONTACT",
    "CONTROLE": "CONTROL",
    "CONTROLO": "CONTROL",
    "COORECTION": "CORRECTION",
    "COPPIED": "COPIED",
    "CORECTION": "CORRECTION",
    "COTACT": "CONTACT",
    "COTINUE": "CONTINUE",
    "COTNACT": "CONTACT",
    "CURCUIT": "CIRCUIT",
    "DEAPARTURE": "DEPARTURE",
    "DEAPRTURE": "DEPARTURE",
    "DECEND": "DESCEND",
    "DEGEES": "DEGREES",
    "DEGRES": "DEGREES",
    "DENCENDING": "DESCENDING",
    "DEPARURE": "DEPARTURE",
    "DESCEDING": "DESCENDING",
    "DESCEN": "DESCEND",
    "DESCENG": "DESCEND",
    "DESCENIDNG": "DESCENDING",
    "DESCNED": "DESCEND",
    "DESECEND": "DESCEND",
    "DESEND": "DESCEND",
    "DESSCEND": "DESCEND",
    "DIREC": "DIRECT",
    "DISCRTION": "DISCRETION",
    "EADING": "HEADING",
    "ESTABLSIH": "ESTABLISH",
    "ESTALBISHED": "ESTABLISHED",
    "ETABLISHED": "ESTABLISHED",
    "ETIOPIAN": "ETHIOPIAN",
    "EVNING": "EVENING",
    "EXEPECT": "EXPECT",
    "EXPERIANCING": "EXPERIENCING",
    "EXTANSION": "EXTENSION",
    "FAVOUR": "FAVOR",
    "FINNARI": "FINNAIR",
    "FLIGTH": "FLIGHT",
    "FOLOW": "FOLLOW",
    "FOURTY": "FORTY",
    "GERMANWING": "GERMANWINGS",
    "GOAHEAD": "GO AHEAD",
    "GODD": "GOOD",
    "GOODBYE": "GOOD BYE",
    "GROSJET": "GROSSJET",
    "GROUDN": "GROUND",
    "HALLO": "HELLO",
    "HEADINT": "HEADING",
    "HEADNIG": "HEADING",
    "HEDING": "HEADING",
    "HODLING": "HOLDING",
    "HUDRED": "HUNDRED",
    "IFORMATION": "INFORMATION",
    "INBOUD": "INBOUND",
    "INBOUDN": "INBOUND",
    "INFOMRATION": "INFORMATION",
    "INITIALY": "INITIALLY",
    "INTERESCTION": "INTERSECTION",
    "KDNOTS": "KNOTS",
    "KNTOS": "KNOTS",
    "LANDA": "LAND",
    "LCIMB": "CLIMB",
    "LENGHT": "LENGTH",
    "LENGT": "LENGTH",
    "LEVELED": "LEVEL",
    "LEVLE": "LEVEL",
    "LIGHER": "LIGHTER",
    "LOUND": "LOUD",
    "LUFHANSA": "LUFTHANSA",
    "LUFHTANSA": "LUFTHANSA",
    "LUFTAHNSA": "LUFTHANSA",
    "LUFTHASNA": "LUFTHANSA",
    "MAINATINANING": "MAINTAINING",
    "MAINTAING": "MAINTAINING",
    "MAINTANING": "MAINTAINING",
    "MAITAIN": "MAINTAIN",
    "MINTUES": "MINUTES",
    "MOLDAVA": "MOLDOVA",
    "MOORNING": "MORNING",
    "NEAGATIVE": "NEGATIVE",
    "NINTEEN": "NINETEEN",
    "NINTY": "NINETY",
    "NOICE": "NOISE",
    "NORTHSHUTTLE": "NORSHUTTLE",
    "NORTHSTHUTTEL": "NORSHUTTLE",
    "NORTHSTHUTTLE": "NORSHUTTLE",
    "NOSIG": "NOSING",
    "NOSRHUTLE": "NORSHUTTLE",
    "OPOSITE": "OPPOSITE",
    "OT": "TO",
    "PASSINF": "PASSING",
    "PASSIN": "PASSING",
    "PLESE": "PLEASE",
    "POSSBILE": "POSSIBLE",
    "PREFERED": "PREFERRED",
    "PROCCEDING": "PROCEEDING",
    "PROCEEDTO": "PROCEED TO",
    "PSSING": "PASSING",
    "QHN": "QNH",
    "QUANTAS": "QANTAS",
    "QUATARI": "QATARI",
    "RADR": "RADAR",
    "READBACK": "READ BACK",
    "RECOMEND": "RECOMMEND",
    "RECOMEND": "RECOMMEND",
    "REQEUSTED": "REQUESTED",
    "REQEUST": "REQUEST",
    "REQUESTE": "REQUEST",
    "REQUSTED": "REQUESTED",
    "REQUSTING": "REQUESTING",
    "RESETING": "RESETTING",
    "RESRTICTION": "RESTRICTION",
    "RESTRCTIONS": "RESTRICTIONS",
    "RESTRISCTION": "RESTRICTION",
    "RIGH": "RIGHT",
    "ROGGER": "ROGER",
    "ROGRE": "ROGER",
    "SESION": "DECISION",
    "SHOTRCUT": "SHORTCUT",
    "SINAGAPORE": "SINGAPORE",
    "SINGAPOOR": "SINGAPORE",
    "SKYRAVEL": "SKYTRAVEL",
    "SKYTAVEL": "SKYTRAVEL",
    "SMARTWING": "SMARTWINGS",
    "SPEEDBIRG": "SPEEDBIRD",
    "SQUAKING": "SQUAWKING",
    "SQUAK": "SQUAWK",
    "SQUWAK": "SQUAWK",
    "STANDAR": "STANDARD",
    "STANDART": "STANDARD",
    "STARTUP": "START UP",
    "SUFFICIAN": "SUFFICIENT",
    "SWTICHING": "SWITCHING",
    "TAHNK": "THANK",
    "TECHNICAN": "TECHNICIAN",
    "TELAVIV": "TEL AVIV",
    "THAT'T": "THAT'S",
    "THIRDY": "THIRTY",
    "THOSUAND": "THOUSAND",
    "THOUASAND": "THOUSAND",
    "TIMECHECK": "TIME CHECK",
    "TRAFIC": "TRAFFIC",
    "TRESHOLD": "THRESHOLD",
    "TUBULENCE": "TURBULENCE",
    "TURBOLENCE": "TURBULENCE",
    "TURUBLENCE": "TURBULENCE",
    "UNREADEBLE": "UNREADABLE",
    "UNTILL": "UNTIL",
    "UTNIL": "UNTIL",
    "VACAT": "VACATE",
    "VECTORIN": "VECTOR IN",
    "WCHICH": "WHICH",
    "WIHT": "WITH",
    "WINE": "WIEN",
    "WIZZIAR": "WIZZAIR",
    "WONDREFUL": "WONDERFUL",
}

COLLAPSE_WORDS = (
    ("AIR SPACE", "AIRSPACE"),
    ("CLEAR FOR", "CLEARED FOR"),
    ("DESCENT TO", "DESCEND TO"),
    ("DESCENT FLIGHT", "DESCEND FLIGHT"),
    ("DESCEND RATE", "DESCENT RATE"),
    ("STAND BYE", "STANDBY"),
)


def text_normalize(
    text: str,
    silence_sym: str,
    breath_sym: str,
    noise_sym: str,
    foreign_sym: str,  # When None, will output foreign words
    unintelligble_sym: str,  # When None, will output unintellible words
    partial_sym: str,  # When None, will output partial words
    unknown_sym: str,
):

    assert is_module_available(
        "num2words"
    ), "Please run 'pip install num2words' for number to word normalization."
    from num2words import num2words

    # regex patterns
    BRACKET_PADDING_PATTERN1 = re.compile(r"([\w\.\+])(\[|\()")
    BRACKET_PADDING_PATTERN2 = re.compile(r"(\]|\))([\w\+])")
    COMMENT_PATTERN = re.compile(r"\[comment_\|].*?\[\|_comment]")
    BACKGROUND_SPEECH_PATTERN = re.compile(
        r"\[background_speech_\|](.*?)\[\|_background_speech]"
    )
    NOISE_PATTERN = re.compile(r"\[noise_\|](.*?)\[\|_noise]")
    SPEAKER_PATTERN = re.compile(r"\[speaker_\|](.*?)\[\|_speaker]")
    DECIMAL_NUMBER_PATTERN = re.compile(r"\.([0-9])")
    NUMBER_DECIMAL_PATTERN = re.compile(r"([0-9])\.")
    PHONETIC_INTERRUPTED_PATTERN1 = re.compile(r"([A-Z]+\+)")
    PHONETIC_INTERRUPTED_PATTERN2 = re.compile(r"(\+[A-Z]+)")
    INTERRUPTED_PATTERN1 = re.compile(r"(\w+\+)")
    INTERRUPTED_PATTERN2 = re.compile(r"(\+\w+)")
    ABBREVIATION_PATTERN = re.compile(r"\(((\w*|\s*|\+)*)\(((\w*|\s*)*)\)\)")
    SPLIT_NUMERIC_ALPHA = re.compile(r"([0-9])([A-Za-z])")
    SPLIT_ALPHA_NUMERIC = re.compile(r"([A-Za-z])([0-9])")
    NO_ENG_PATTERN = re.compile(r"\[NO_ENG_\|](.*?)\[\|_NO_ENG]")
    CZECH_PATTERN = re.compile(r"\[CZECH_\|](.*?)\[\|_CZECH]")
    UNINTELLIGIBLE_PATTERN = re.compile(
        r"\[UNINTELLIGIBLE_\|](.*?)\[\|_UNINTELLIGIBLE]"
    )
    WHITESPACE_PATTERN = re.compile(r"  +")

    text = BRACKET_PADDING_PATTERN1.sub(r"\1 \2", text)
    text = BRACKET_PADDING_PATTERN2.sub(r"\1 \2", text)
    text = text.replace("](", "] (")

    text = text.replace("°", "")
    text = text.replace("?", "")
    text = text.replace("¨", "")
    text = text.replace("´", "'")

    text = COMMENT_PATTERN.sub("", text)

    text = BACKGROUND_SPEECH_PATTERN.sub(r"\1", text)
    text = NOISE_PATTERN.sub(r"\1", text)
    text = SPEAKER_PATTERN.sub(r"\1", text)
    text = DECIMAL_NUMBER_PATTERN.sub(r". \1", text)
    text = NUMBER_DECIMAL_PATTERN.sub(r"\1 .", text)

    text = PHONETIC_INTERRUPTED_PATTERN1.sub(lambda m: m.group(1).lower(), text)
    text = PHONETIC_INTERRUPTED_PATTERN2.sub(lambda m: m.group(1).lower(), text)

    text = ABBREVIATION_PATTERN.sub(r"\1", text)

    # apply this word fix before applying split numeric-alpha patterns
    text = text.replace("6raha", "praha")

    text = SPLIT_NUMERIC_ALPHA.sub(r"\1 \2", text)
    text = SPLIT_ALPHA_NUMERIC.sub(r"\1 \2", text)

    text = strip_accents(text)

    simple_replace = {
        "[ehm_]": breath_sym,
        "[noise]": noise_sym,
        "[unintelligible]": unknown_sym,
        "[background_speech]": noise_sym,
        "[speaker]": breath_sym,
    }

    result = []
    for w in text.split():
        if w in simple_replace:
            result.append(simple_replace[w])
        elif w in UNKNOWN_ABBREVIATIONS:
            result.append(unknown_sym)
        elif w in ABBREVIATIONS:
            result.append(ABBREVIATIONS[w])
        elif w in INDIVIDUALLY_PRONOUNCED:
            result.append(" ".join([*w]).upper())
        elif w in PHONETIC_ALPHABET:
            result.append(PHONETIC_ALPHABET[w])
        elif w.isdigit():
            result.append(num2words(w).replace("-", " ").replace(",", "").upper())
        else:
            result.append(w.upper())
    text = " ".join(result)

    # from here on, text, with the exception of inserted symbols, is all uppercase,
    # therfore [no_eng], [czech] and [unintelligible] pattern matching must be done in uppercase.

    if foreign_sym == None:
        text = text.replace("[NO_ENG]", unknown_sym)
        text = NO_ENG_PATTERN.sub(r"\1", text)
        text = CZECH_PATTERN.sub(r"\1", text)
    else:
        text = text.replace("[NO_ENG]", foreign_sym)
        text = NO_ENG_PATTERN.sub(foreign_sym, text)
        text = CZECH_PATTERN.sub(foreign_sym, text)

    if unintelligble_sym == None:
        text = UNINTELLIGIBLE_PATTERN.sub(r"\1", text)
    else:
        text = UNINTELLIGIBLE_PATTERN.sub(unintelligble_sym, text)

    if partial_sym != None:
        text = INTERRUPTED_PATTERN1.sub(partial_sym, text)
        text = INTERRUPTED_PATTERN2.sub(partial_sym, text)

    text = text.replace("+", "")

    text = WHITESPACE_PATTERN.sub(" ", text)
    text = text.strip()

    text = " ".join([FIX_TYPOS[w] if w in FIX_TYPOS else w for w in text.split()])

    for pair in COLLAPSE_WORDS:
        text = text.replace(pair[0], pair[1])

    return text


SPEAKER_TO_ID_SUFFIX = {"air_ground": "PIAT", "ground": "AT", "air": "PI"}


def finish_segment(supervisions, segment, end_time):
    segment.duration = end_time - segment.start
    segment.id += "_%06d_%s" % (end_time * 100, SPEAKER_TO_ID_SUFFIX[segment.speaker])
    supervisions.append(segment)


def prepare_uwb_atcc(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    silence_sym: Optional[str] = "",
    breath_sym: Optional[str] = "",
    noise_sym: Optional[str] = "",
    foreign_sym: Optional[str] = "<unk>",
    partial_sym: Optional[str] = "<unk>",
    unintelligble_sym: Optional[str] = "<unk>",
    unknown_sym: Optional[str] = "<unk>",
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param silence_sym: str, silence symbol
    :param breath_sym: str, breath symbol
    :param noise_sym: str, noise symbol
    :param foreign_sym: str, foreign symbol. when set to None, will output foreign words
    :param partial_sym: str, partial symbol. When set to None, will output partial words
    :param unintelligble_sym: str, unintellible symbol. When set to None, will output unintelligble words
    :param unknown_sym: str, unknown symbol
    :return: The RecordingSet and SupervisionSet with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    trs_files = sorted(corpus_dir.glob("*.trs"), key=lambda p: p.name)
    assert len(trs_files) == 2657

    recordings = []
    supervisions = []

    # regex pattern for multiple whitespaces
    WHITESPACE_PATTERN = re.compile(r"  +")

    from tqdm.auto import tqdm

    for t in tqdm(trs_files, desc="Preparing"):
        # repair broken xml files
        if t.stem in (
            "ACCU-80UXVV",
            "ACCU-7NqzYv",
            "ACCU-PhR5Oj",
            "ACCU-JaeNLH",
            "TWR-XgqNSk",
        ):
            with open(t, encoding="cp1250") as f:
                root = ET.fromstring(f.read() + "</Turn></Section></Episode></Trans>")
        else:
            root = ET.parse(t).getroot()

        audio_path = corpus_dir / root.attrib["audio_filename"][len("e2_") :]
        if not audio_path.is_file():
            logging.warning(f"No such file: {audio_path}")
            continue

        recording = Recording.from_file(audio_path)
        recordings.append(recording)

        last_segment = None

        for section in root.findall(".//Section"):
            for turn in section:
                if turn.tag != "Turn":
                    logging.warning(f"Unexpected tag: {turn.tag}")
                    continue

                end_time = float(turn.attrib["endTime"])

                for sync in turn:
                    if sync.tag != "Sync":
                        logging.warning(f"Unexpected tag: {sync.tag}")
                        continue

                    time = float(sync.attrib["time"])

                    if last_segment:
                        finish_segment(supervisions, last_segment, time)
                        last_segment = None

                    text = sync.tail.strip()
                    if text == "":
                        continue

                    orig_text = text

                    # extract speaker id
                    if "[air_|]" in text or "[ground_|]" in text:
                        speaker = "air_ground"
                    elif "[air]" in text:
                        speaker = "air"
                    elif "[ground]" in text:
                        speaker = "ground"
                    else:
                        continue

                    # remove speaker id tags
                    text = text.replace("][", "] [")
                    for label in (
                        "[air_|]",
                        "[|_air]",
                        "[ground_|]",
                        "[|_ground]",
                        "[air]",
                        "[ground]",
                    ):
                        text = text.replace(label, "")

                    text = text_normalize(
                        text,
                        silence_sym=silence_sym,
                        breath_sym=breath_sym,
                        noise_sym=noise_sym,
                        foreign_sym=foreign_sym,
                        partial_sym=partial_sym,
                        unintelligble_sym=unintelligble_sym,
                        unknown_sym=unknown_sym,
                    )

                    if text == "":
                        continue

                    last_segment = SupervisionSegment(
                        id="uwb-atcc_%s_%06d" % (audio_path.stem, time * 100),
                        recording_id=recording.id,
                        start=time,
                        duration=0,  # populated by finish_segment()
                        channel=0,
                        language="English",
                        text=text,
                        speaker=speaker,
                        custom={
                            "type": section.attrib["type"],
                            "orig_text": WHITESPACE_PATTERN.sub(" ", orig_text.strip()),
                        },
                    )

                if last_segment:
                    finish_segment(supervisions, last_segment, end_time)
                    last_segment = None

    recording_set = RecordingSet.from_recordings(recordings)
    supervision_set = SupervisionSet.from_segments(supervisions)

    recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
    validate_recordings_and_supervisions(recording_set, supervision_set)

    if output_dir is not None:
        supervision_set.to_file(output_dir / "uwb_atcc_supervisions_all.jsonl.gz")
        recording_set.to_file(output_dir / "uwb_atcc_recordings_all.jsonl.gz")

    return {"recordings": recording_set, "supervisions": supervision_set}



from dataclasses import dataclass, field
from typing import List, Optional

import numpy as np
import torch

try:
    # Pytorch >= 1.7
    from torch.fft import irfft, rfft
except ImportError:
    from torch import irfft, rfft

# Implementation based on torch-audiomentations:
# https://github.com/asteroid-team/torch-audiomentations/blob/master/torch_audiomentations/utils/convolution.py

_NEXT_FAST_LEN = {}


def next_fast_len(size):
    """
    Returns the next largest number ``n >= size`` whose prime factors are all
    2, 3, or 5. These sizes are efficient for fast fourier transforms.
    Equivalent to :func:`scipy.fftpack.next_fast_len`.
    Note: This function was originally copied from the https://github.com/pyro-ppl/pyro
    repository, where the license was Apache 2.0. Any modifications to the original code can be
    found at https://github.com/asteroid-team/torch-audiomentations/commits
    :param int size: A positive number.
    :returns: A possibly larger number.
    :rtype int:
    """
    try:
        return _NEXT_FAST_LEN[size]
    except KeyError:
        pass

    assert isinstance(size, int) and size > 0
    next_size = size
    while True:
        remaining = next_size
        for n in (2, 3, 5):
            while remaining % n == 0:
                remaining //= n
        if remaining == 1:
            _NEXT_FAST_LEN[size] = next_size
            return next_size
        next_size += 1


def convolve1d(signal: torch.Tensor, kernel: torch.Tensor) -> torch.Tensor:
    """
    Computes the 1-d convolution of signal by kernel using FFTs.
    Both signal and kernel must be 1-dimensional.
    :param torch.Tensor signal: A signal to convolve.
    :param torch.Tensor kernel: A convolution kernel.
    :param str mode: One of: 'full', 'valid', 'same'.
    :return: torch.Tensor Convolution of signal with kernel. Returns the full convolution, i.e.,
        the output tensor will have size m + n - 1, where m is the length of the
        signal and n is the length of the kernel.
    """
    assert (
        signal.ndim == 1 and kernel.ndim == 1
    ), "signal and kernel must be 1-dimensional"
    m = signal.size(-1)
    n = kernel.size(-1)

    # Compute convolution using fft.
    padded_size = m + n - 1
    # Round up for cheaper fft.
    fast_ftt_size = next_fast_len(padded_size)
    f_signal = rfft(signal, n=fast_ftt_size)
    f_kernel = rfft(kernel, n=fast_ftt_size)
    f_result = f_signal * f_kernel
    result = irfft(f_result, n=fast_ftt_size)

    return result[:padded_size]


# The following is based on: https://github.com/yluo42/FRA-RIR/blob/main/FRA-RIR.py
@dataclass
class FastRandomRIRGenerator:
    sr: int = 16000
    direct_range: List = field(default_factory=lambda: [-6, 50])
    max_T60: float = 0.8
    alpha: float = 0.25
    a: float = -2.0
    b: float = 2.0
    tau: float = 0.2
    room_seed: Optional[int] = None
    source_seed: Optional[int] = None

    def __post_init__(self):
        self.room_rng = (
            np.random.default_rng(self.room_seed)
            if self.room_seed is not None
            else np.random.default_rng()
        )
        self.source_rng = (
            np.random.default_rng(self.source_seed)
            if self.source_seed is not None
            else np.random.default_rng()
        )

    def __call__(self, nsource: int = 1) -> np.ndarray:
        """
        :param nsource: number of sources (RIR filters) to simulate. Default: 1.
        :return: simulated RIR filter for all sources, shape: (nsource, nsample)
        """
        from lhotse.augmentation.torchaudio import (
            check_for_torchaudio,
            get_or_create_resampler,
        )

        check_for_torchaudio()

        from torchaudio.functional import highpass_biquad

        # the sample rate at which the original RIR filter is generated
        ratio = 64
        sample_sr = self.sr * ratio

        # two resampling operations
        resample1 = get_or_create_resampler(sample_sr, sample_sr // int(np.sqrt(ratio)))
        resample2 = get_or_create_resampler(sample_sr // int(np.sqrt(ratio)), self.sr)

        eps = np.finfo(np.float16).eps

        # sample T60 of the room
        T60 = torch.from_numpy(self.room_rng.uniform(0.1, self.max_T60, size=(1,)))[
            0
        ].data

        # sample room-related statistics for calculating the reflection coefficient R
        R = torch.from_numpy(self.room_rng.uniform(0.1, 1.2, size=(1,)))[0].data

        # sample distance between the sound sources and the receiver (d_0) if not given
        direct_dist = torch.from_numpy(
            self.source_rng.uniform(0.2, 12.0, size=(nsource,))
        )

        # number of virtual sound sources
        image = self.sr * 2

        # sound velocity
        velocity = 340.0

        # indices of direct-path signals based on the sampled d_0
        direct_idx = torch.ceil(direct_dist * sample_sr / velocity).long()

        # length of the RIR filter based on the sampled T60
        rir_length = int(np.ceil(sample_sr * T60))

        # calculate the reflection coefficient based on the Eyring's empirical equation
        reflect_coef = (1 - (1 - torch.exp(-0.16 * R / T60)).pow(2)).sqrt()

        # randomly sample the propagation distance for all the virtual sound sources
        dist_range = [
            torch.linspace(1.0, velocity * T60 / direct_dist[i] - 1, image)
            for i in range(nsource)
        ]
        # a simple quadratic function
        dist_prob = torch.linspace(self.alpha, 1.0, image).pow(2)
        dist_prob = dist_prob / dist_prob.sum()
        dist_select_idx = dist_prob.multinomial(
            num_samples=image * nsource, replacement=True
        ).view(nsource, image)
        # the distance is sampled as a ratio between d_0 and each virtual sound sources
        dist_ratio = torch.stack(
            [dist_range[i][dist_select_idx[i]] for i in range(nsource)], 0
        )
        dist = direct_dist.view(-1, 1) * dist_ratio

        # sample the number of reflections (can be nonintegers)
        # calculate the maximum number of reflections
        reflect_max = (
            torch.log10(velocity * T60) - torch.log10(direct_dist) - 3
        ) / torch.log10(reflect_coef + eps)
        # calculate the number of reflections based on the assumption that
        # virtual sound sources which have longer propagation distances may reflect more frequently
        reflect_ratio = (dist / (velocity * T60)).pow(2) * (
            reflect_max.view(nsource, -1) - 1
        ) + 1
        # add a random pertubation based on the assumption that
        # virtual sound sources which have similar propagation distances can have different routes and reflection patterns
        reflect_pertub = torch.from_numpy(
            self.source_rng.uniform(self.a, self.b, size=(nsource, image))
        ) * (dist_ratio.pow(self.tau))
        # all virtual sound sources should reflect for at least once
        reflect_ratio = torch.maximum(reflect_ratio + reflect_pertub, torch.ones(1))

        # calculate the rescaled dirac comb as RIR filter
        dist = torch.cat([direct_dist.reshape(-1, 1), dist], 1)
        reflect_ratio = torch.cat([torch.zeros(nsource, 1), reflect_ratio], 1)
        rir = torch.zeros(nsource, rir_length)
        delta_idx = torch.minimum(
            torch.ceil(dist * sample_sr / velocity), torch.ones(1) * rir_length - 1
        ).long()
        delta_decay = reflect_coef.pow(reflect_ratio) / dist
        for i in range(nsource):
            rir[i][delta_idx[i]] += delta_decay[i]

        # a binary mask for direct-path RIR
        direct_mask = torch.zeros(nsource, rir_length).float()
        for i in range(nsource):
            direct_mask[
                i,
                max(direct_idx[i] + sample_sr * self.direct_range[0] // 1000, 0) : min(
                    direct_idx[i] + sample_sr * self.direct_range[1] // 1000, rir_length
                ),
            ] = 1.0
        rir_direct = rir * direct_mask

        # downsample
        all_rir = torch.stack([rir, rir_direct], 1).view(nsource * 2, -1)
        rir_downsample = resample1(all_rir)

        # apply high-pass filter
        rir_hp = highpass_biquad(rir_downsample, sample_sr // int(np.sqrt(ratio)), 80.0)

        # downsample again
        rir = resample2(rir_hp).float().view(nsource, 2, -1)

        # RIR filter and direct-path RIR filter at target sample rate
        rir_filter = rir[:, 0]  # nsource, nsample

        # convert to numpy array
        return rir_filter.numpy()



import logging
import warnings
from dataclasses import dataclass
from functools import partial, reduce
from operator import add
from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch

from lhotse.audio import Recording
from lhotse.cut.data import DataCut
from lhotse.features import Features
from lhotse.supervision import SupervisionSegment
from lhotse.utils import (
    add_durations,
    fastcopy,
    hash_str_to_int,
    is_equal_or_contains,
    merge_items_with_delimiter,
    overlaps,
    rich_exception_info,
    uuid4,
)


@dataclass
class MonoCut(DataCut):
    """
    :class:`~lhotse.cut.MonoCut` is a :class:`~lhotse.cut.Cut` of a single channel of
    a :class:`~lhotse.audio.Recording`. In addition to Cut, it has a specified channel attribute. This is the most commonly used type of cut.

    Please refer to the documentation of :class:`~lhotse.cut.Cut` to learn more about using cuts.

    See also:

        - :class:`lhotse.cut.Cut`
        - :class:`lhotse.cut.MixedCut`
        - :class:`lhotse.cut.CutSet`
    """

    channel: int

    @property
    def num_channels(self) -> int:
        # MonoCut is always single-channel.
        return 1

    @rich_exception_info
    def load_features(self) -> Optional[np.ndarray]:
        """
        Load the features from the underlying storage and cut them to the relevant
        [begin, duration] region of the current MonoCut.
        """
        if self.has_features:
            feats = self.features.load(start=self.start, duration=self.duration)
            # Note: we forgive off-by-one errors in the feature matrix frames
            #       due to various hard-to-predict floating point arithmetic issues.
            #       If needed, we will remove or duplicate the last frame to be
            #       consistent with the manifests declared "num_frames".
            if feats.shape[0] - self.num_frames == 1:
                feats = feats[: self.num_frames, :]
            elif feats.shape[0] - self.num_frames == -1:
                feats = np.concatenate((feats, feats[-1:, :]), axis=0)
            return feats
        return None

    @rich_exception_info
    def load_audio(self) -> Optional[np.ndarray]:
        """
        Load the audio by locating the appropriate recording in the supplied RecordingSet.
        The audio is trimmed to the [begin, end] range specified by the MonoCut.

        :return: a numpy ndarray with audio samples, with shape (1 <channel>, N <samples>)
        """
        if self.has_recording:
            return self.recording.load_audio(
                channels=self.channel,
                offset=self.start,
                duration=self.duration,
            )
        return None

    @rich_exception_info
    def load_video(
        self,
        with_audio: bool = True,
    ) -> Optional[Tuple[torch.Tensor, Optional[torch.Tensor]]]:
        """
        Load the subset of video (and audio) from attached recording.
        The data is trimmed to the [begin, end] range specified by the MonoCut.

        :param with_audio: bool, whether to load and return audio alongside video. True by default.
        :return: a tuple of video tensor and optionally audio tensor (or ``None``),
            or ``None`` if this cut has no video.
        """
        if self.has_video:
            return self.recording.load_video(
                channels=self.channel,
                offset=self.start,
                duration=self.duration,
                with_audio=with_audio,
            )
        return None

    def with_channels(self, channels: Union[List[int], int]) -> DataCut:
        """
        Select specified channels from this cut.
        Supports extending to other channels available in the underlying :class:`Recording`.
        If a single channel is provided, we'll return a :class:`~lhotse.cut.MonoCut`,
        otherwise we'll return a :class:`~lhotse.cut.MultiCut`.
        """
        channel_is_int = isinstance(channels, int)
        assert set([channels] if channel_is_int else channels).issubset(
            set(self.recording.channel_ids)
        ), f"Cannot select {channels=} because they are not a subset of {self.recording.channel_ids=}"

        mono = channel_is_int or len(channels) == 1
        if mono:
            if not channel_is_int:
                (channels,) = channels
            return MonoCut(
                id=f"{self.id}-{channels}",
                recording=self.recording,
                start=self.start,
                duration=self.duration,
                channel=channels,
                supervisions=[
                    fastcopy(s, channel=channels)
                    for s in self.supervisions
                    if is_equal_or_contains(s.channel, channels)
                ],
                custom=self.custom,
            )
        else:
            from lhotse import MultiCut

            return MultiCut(
                id=f"{self.id}-{len(channels)}chan",
                start=self.start,
                duration=self.duration,
                channel=channels,
                supervisions=[
                    s
                    for s in self.supervisions
                    if is_equal_or_contains(channels, s.channel)
                ],
                custom=self.custom,
            )

    def reverb_rir(
        self,
        rir_recording: Optional[Union[Recording, DataCut]] = None,
        normalize_output: bool = True,
        early_only: bool = False,
        affix_id: bool = True,
        rir_channels: Sequence[int] = (0,),
        room_rng_seed: Optional[int] = None,
        source_rng_seed: Optional[int] = None,
    ) -> DataCut:
        """
        Return a new ``DataCut`` that will convolve the audio with the provided impulse response.
        If the `rir_recording` is multi-channel, the `rir_channels` argument determines which channels
        will be used. By default, we use the first channel and return a MonoCut. If we reverberate
        with a multi-channel RIR, we return a MultiCut.

        If no ``rir_recording`` is provided, we will generate an impulse response using a fast random
        generator (https://arxiv.org/abs/2208.04101). Note that the generator only supports simulating
        reverberation with a single microphone, so we will return a MonoCut in this case.

        :param rir_recording: The impulse response to use for convolving.
        :param normalize_output: When true, output will be normalized to have energy as input.
        :param early_only: When true, only the early reflections (first 50 ms) will be used.
        :param affix_id: When true, we will modify the ``MonoCut.id`` field
            by affixing it with "_rvb".
        :param rir_channels: The channels of the impulse response to use. First channel is used by default.
            If multiple channels are specified, this will produce a MultiCut instead of a MonoCut.
        :param room_rng_seed: The seed for the room configuration.
        :param source_rng_seed: The seed for the source position.
        :return: a modified copy of the current ``MonoCut``.
        """
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot apply reverberation on a MonoCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to reverberate a MonoCut that references pre-computed features. "
                "The feature manifest will be detached, as we do not support feature-domain "
                "reverberation."
            )
            self.features = None

        assert rir_recording is None or all(
            c < rir_recording.num_channels for c in rir_channels
        ), "Invalid channel index in `rir_channels`."

        if rir_recording is None:
            # Set rir_channels to 0 since we can only generate a single-channel RIR.
            rir_channels = [0]

            if room_rng_seed is None:
                room_rng_seed = hash_str_to_int(str(uuid4()) + self.id)

            if source_rng_seed is None:
                source_rng_seed = room_rng_seed

        if len(rir_channels) == 1:
            # reverberation will return a MonoCut
            recording_rvb = self.recording.reverb_rir(
                rir_recording=rir_recording,
                normalize_output=normalize_output,
                early_only=early_only,
                affix_id=affix_id,
                rir_channels=rir_channels,
                room_rng_seed=room_rng_seed,
                source_rng_seed=source_rng_seed,
            )
            # Match the supervision's id (and it's underlying recording id).
            supervisions_rvb = [
                s.reverb_rir(
                    affix_id=affix_id,
                )
                for s in self.supervisions
            ]

            return fastcopy(
                self,
                id=f"{self.id}_rvb" if affix_id else self.id,
                recording=recording_rvb,
                supervisions=supervisions_rvb,
            )
        else:
            from .multi import MultiCut

            channels = list(range(len(rir_channels)))
            # we will return a MultiCut where each channel represents the MonoCut convolved
            # with a single channel of the RIR
            recording_rvb = self.recording.reverb_rir(
                rir_recording=rir_recording,
                normalize_output=normalize_output,
                early_only=early_only,
                affix_id=affix_id,
                rir_channels=rir_channels,
                room_rng_seed=room_rng_seed,
                source_rng_seed=source_rng_seed,
            )
            # Match the supervision's id (and it's underlying recording id).
            supervisions_rvb = [
                s.reverb_rir(
                    affix_id=affix_id,
                    channel=channels,
                )
                for s in self.supervisions
            ]

            return fastcopy(
                MultiCut.from_mono(self),
                recording=recording_rvb,
                supervisions=supervisions_rvb,
                channel=channels,
            )

    def merge_supervisions(
        self,
        merge_policy: str = "delimiter",
        custom_merge_fn: Optional[Callable[[str, Iterable[Any]], Any]] = None,
    ) -> "MonoCut":
        """
        Return a copy of the cut that has all of its supervisions merged into
        a single segment.

        The new start is the start of the earliest superivion, and the new duration
        is a minimum spanning duration for all the supervisions. The text fields of
        all segments are concatenated with a whitespace.

        :param merge_policy: one of "keep_first" or "delimiter". If "keep_first", we
            keep only the first segment's field value, otherwise all string fields
            (including IDs) are prefixed with "cat#" and concatenated with a hash symbol "#".
            This is also applied to ``custom`` fields. Fields with a ``None`` value are omitted.
        :param custom_merge_fn: a function that will be called to merge custom fields values.
            We expect ``custom_merge_fn`` to handle all possible custom keys.
            When not provided, we will treat all custom values as strings.
            It will be called roughly like:
            ``custom_merge_fn(custom_key, [s.custom[custom_key] for s in sups])``
        """
        merge_func_ = partial(
            merge_items_with_delimiter,
            delimiter="#",
            return_first=(merge_policy == "keep_first"),
        )

        # "m" stands for merged in variable names below

        if custom_merge_fn is not None:
            # Merge custom fields with the user-provided function.
            merge_custom = custom_merge_fn
        else:
            # Merge the string representations of custom fields.
            merge_custom = lambda k, vs: merge_func_(map(str, vs))

        sups = sorted(self.supervisions, key=lambda s: s.start)

        if len(sups) <= 1:
            return self

        # the sampling rate is arbitrary, ensures there are no float precision errors
        mstart = sups[0].start
        mend = sups[-1].end
        mduration = add_durations(mend, -mstart, sampling_rate=self.sampling_rate)

        custom_keys = set(
            k for s in sups if s.custom is not None for k in s.custom.keys()
        )
        alignment_keys = set(
            k for s in sups if s.alignment is not None for k in s.alignment.keys()
        )

        if any(overlaps(s1, s2) for s1, s2 in zip(sups, sups[1:])) and any(
            s.text is not None for s in sups
        ):
            warnings.warn(
                "You are merging overlapping supervisions that have text transcripts. "
                "The result is likely to be unusable if you are going to train speech "
                f"recognition models (cut id: {self.id})."
            )

        msup = SupervisionSegment(
            id=merge_func_(s.id for s in sups),
            recording_id=sups[0].recording_id,
            start=mstart,
            duration=mduration,
            channel=sups[0].channel,
            text=" ".join(s.text for s in sups if s.text),
            speaker=merge_func_(s.speaker for s in sups if s.speaker),
            language=merge_func_(s.language for s in sups if s.language),
            gender=merge_func_(s.gender for s in sups if s.gender),
            custom={
                k: merge_custom(
                    k,
                    (
                        s.custom[k]
                        for s in sups
                        if s.custom is not None and k in s.custom
                    ),
                )
                for k in custom_keys
            },
            alignment={
                # Concatenate the lists of alignment units.
                k: reduce(
                    add,
                    (
                        s.alignment[k]
                        for s in sups
                        if s.alignment is not None and k in s.alignment
                    ),
                )
                for k in alignment_keys
            },
        )

        return fastcopy(self, supervisions=[msup])

    @staticmethod
    def from_dict(data: dict) -> "MonoCut":
        from lhotse.serialization import deserialize_custom_field

        # Remove "type" field if exists.
        data.pop("type", None)

        features = (
            Features.from_dict(data.pop("features")) if "features" in data else None
        )
        recording = (
            Recording.from_dict(data.pop("recording")) if "recording" in data else None
        )
        supervision_infos = data.pop("supervisions") if "supervisions" in data else []

        if "custom" in data:
            deserialize_custom_field(data["custom"])

        if "type" in data:
            data.pop("type")

        return MonoCut(
            **data,
            features=features,
            recording=recording,
            supervisions=[SupervisionSegment.from_dict(s) for s in supervision_infos],
        )



import warnings
from typing import Any, Dict, List, Optional

from lhotse import CutSet, Seconds
from lhotse.dataset.sampling.base import TimeConstraint
from lhotse.dataset.sampling.data_source import WeightedDataSource
from lhotse.dataset.sampling.simple import SimpleCutSampler


class WeightedSimpleCutSampler(SimpleCutSampler):
    """
    Samples cuts from a CutSet, where the sampling prob is given by a list.
    To enable global sampling, cuts must be in eager mode.

    When performing sampling, it avoids having duplicated cuts in the same batch.
    The sampler terminates if the number of sampled cuts reach :attr:`num_samples`

    When one of :attr:`max_frames`, :attr:`max_samples`, or :attr:`max_duration` is specified,
    the batch size is dynamic.

    Example usage:

        >>> dataset = K2SpeechRecognitionDataset(cuts)
        >>> weights = get_weights(cuts)
        >>> sampler = WeightedSimpleCutSampler(cuts, weights, num_samples=100, max_duration=200.0)
        >>> loader = DataLoader(dataset, sampler=sampler, batch_size=None)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(
        self,
        cuts: CutSet,
        cuts_weight: List,
        num_samples: int,
        max_duration: Seconds = None,
        max_cuts: Optional[int] = None,
        shuffle: bool = False,
        drop_last: bool = False,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        seed: int = 0,
    ):
        """
        WeightedSimpleCutSampler's constructor

        :param cuts: the ``CutSet`` to sample data from.
        :param cuts_weight: the weight of each cut for sampling.
        :param num_samples: the number of samples to be drawn.
        :param max_duration: The maximum total recording duration from ``cuts``.
        :param max_cuts: The maximum number of cuts sampled to form a mini-batch.
            By default, this constraint is off.
        :param shuffle: When ``True``, the cuts will be shuffled at the start of iteration.
            Convenient when mini-batch loop is inside an outer epoch-level loop, e.g.:
            `for epoch in range(10): for batch in dataset: ...` as every epoch will see a
            different cuts order.
        :param drop_last: When ``True``, the last batch is dropped if it's incomplete.
        :param world_size: Total number of distributed nodes. We will try to infer it by default.
        :param rank: Index of distributed node. We will try to infer it by default.
        :param seed: Random seed used to consistently shuffle the dataset across different processes.
        """
        super().__init__(
            cuts=cuts,
            drop_last=drop_last,
            shuffle=shuffle,
            world_size=world_size,
            rank=rank,
            max_duration=max_duration,
            max_cuts=max_cuts,
            seed=seed,
        )
        assert not cuts.is_lazy, "This sampler does not support lazy mode!"
        self.data_source = WeightedDataSource(
            cuts, weights=cuts_weight, num_samples=num_samples
        )

        self.weights = cuts_weight
        self.num_samples = num_samples

    def state_dict(self) -> Dict[str, Any]:
        """
        Return the current state of the sampler in a state_dict.
        Together with ``load_state_dict()``, this can be used to restore the
        training loop's state to the one stored in the state_dict.
        """
        state_dict = super().state_dict()
        state_dict.update(
            {
                "time_constraint": self.time_constraint.state_dict(),
                "weights": self.weights,
                "num_samples": self.num_samples,
            }
        )
        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Restore the state of the sampler that is described in a state_dict.
        This will result in the sampler yielding batches from where the previous training left it off.

        .. caution::
            The samplers are expected to be initialized with the same CutSets,
            but this is not explicitly checked anywhere.

        .. caution::
            The input ``state_dict`` is being mutated: we remove each consumed key, and expect
            it to be empty at the end of loading. If you don't want this behavior, pass a copy
            inside of this function (e.g., using ``import deepcopy``).

        .. note::
            For implementers of sub-classes of CutSampler: the flag ``self._just_restored_state`` has to be
            handled in ``__iter__`` to make it avoid resetting the just-restored state (only once).
        """
        time_constraint = TimeConstraint(**state_dict.pop("time_constraint"))
        if self.time_constraint != time_constraint:
            warnings.warn(
                "SimpleCutSampler.load_state_dict(): Inconsistent time_constraint:\n"
                f"expected {self.time_constraint}\n"
                f"received {time_constraint}\n"
                f"We will overwrite the settings with the received state_dict."
            )
        self.time_constraint = time_constraint

        super().load_state_dict(state_dict)

        # Restore the data source's state
        self.data_source.fast_forward(self.diagnostics.current_epoch_stats.total_cuts)

        self.weights = state_dict.pop("weights")
        self.num_samples = state_dict.pop("num_samples")

    def __iter__(self) -> "WeightedSimpleCutSampler":
        """
        Prepare the dataset for iterating over a new epoch. Will shuffle the data if requested.
        """
        # Restored state with load_state_dict()? Skip resetting only this once.
        if self._just_restored_state:
            return self
        # Why reset the current epoch?
        # Either we are iterating the epoch for the first time and it's a no-op,
        # or we are iterating the same epoch again, in which case setting more steps
        # than are actually available per epoch would have broken the checkpoint restoration.
        self.diagnostics.reset_current_epoch()
        # Reset the state to the beginning of the epoch.
        iter(self.data_source)
        return self



import copy
import os
import warnings
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from dataclasses import asdict, dataclass
from math import isclose
from typing import Any, Callable, Dict, Iterable, Literal, Optional, Tuple, Union

import torch
from torch import distributed as dist
from torch.utils.data import Sampler

from lhotse.cut import Cut, CutSet
from lhotse.cut.text import TextExample
from lhotse.lazy import Dillable
from lhotse.manipulation import combine
from lhotse.utils import Seconds, ifnone, is_none_or_gt


class CutSampler(Sampler, Dillable):
    """
    ``CutSampler`` is responsible for collecting batches of cuts, given specified criteria.
    It implements correct handling of distributed sampling in ``DataLoader``,
    so that the cuts are not duplicated across workers.

    Sampling in a ``CutSampler`` is intended to be very quick - it only uses the metadata in
    ``CutSet`` manifest to select the cuts, and is not intended to perform any I/O.

    CutSampler works similarly to PyTorch's DistributedSampler - when :attr:`shuffle=True`,
    you should call ``sampler.set_epoch(epoch)`` at each new epoch to have a different
    ordering of returned elements. However, its actual behaviour is different than that of
    DistributedSampler -- instead of partitioning the underlying cuts into equally sized chunks,
    it will return every N-th batch and skip the other batches (where ``N == world_size``).
    The formula used to determine which batches are returned is:
    ``(batch_idx + (world_size - rank)) % world_size == 0``.
    This ensures that we can return an equal number of batches in all distributed workers
    in spite of using a dynamic batch size, at the cost of skipping at most ``world_size - 1`` batches.

    Example usage::

        >>> dataset = K2SpeechRecognitionDataset(cuts)
        >>> sampler = SimpleCutSampler(cuts, max_duration=200, shuffle=True)
        >>> loader = DataLoader(dataset, sampler=sampler, batch_size=None)
        >>> for epoch in range(start_epoch, n_epochs):
        ...     sampler.set_epoch(epoch)
        ...     train(loader)

    .. note::

        For implementers of new samplers:
        Subclasses of CutSampler are expected to implement ``self._next_batch()`` to introduce specific
        sampling logic (e.g. based on filters such as max number of frames/tokens/etc.).
        CutSampler defines ``__iter__()``, which optionally shuffles the cut IDs, and resets
        ``self.cut_idx`` to zero (to be used and incremented inside of ``_next_batch()``.
    """

    def __init__(
        self,
        shuffle: bool = False,
        drop_last: bool = False,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        seed: Union[int, Literal["randomized", "trng"]] = 0,
    ) -> None:
        """
        :param shuffle: When ``True``, the cuts will be shuffled at the start of iteration.
            Convenient when mini-batch loop is inside an outer epoch-level loop, e.g.:
            `for epoch in range(10): for batch in dataset: ...` as every epoch will see a
            different cuts order.
        :param drop_last: When ``True``, the last batch is dropped if it's incomplete.
        :param world_size: Total number of distributed nodes. We will try to infer it by default.
        :param rank: Index of distributed node. We will try to infer it by default.
        :param seed: Random seed used to consistently shuffle the dataset across different processes.
        """
        super().__init__(
            data_source=None
        )  # the "data_source" arg is not used in Sampler...
        self.drop_last = drop_last
        self.shuffle = shuffle
        self.seed = seed
        self.epoch = 0
        self._diagnostics = SamplingDiagnostics()

        # This flag is used to indicate that we have restored a sampler's state from a state_dict.
        # When it is set, we will ignore the next call to iter(), which would have reset the
        # iteration state. This way we can resume training exactly from where it was left off.
        self._just_restored_state = False

        self._maybe_init_distributed(world_size=world_size, rank=rank)
        # By default, self._filter_fn passes every Cut through.
        self._filter_fn: Callable[[Cut], bool] = _filter_nothing()
        self._transforms = []

    @property
    def diagnostics(self):
        """
        Info on how many cuts / batches were returned or rejected during iteration.

        This property can be overriden by child classes e.g. to merge diagnostics of composite samplers.
        """
        return self._diagnostics

    def _maybe_init_distributed(self, world_size: Optional[int], rank: Optional[int]):
        if world_size is not None:
            assert world_size >= 1
        if rank is not None:
            assert rank >= 0

        # Order of precedence:
        # 1. When world size or rank are explicitly provided, we will use them.
        # 2. Next, check WORLD_SIZE and RANK env variables; yes? use them.
        # 3. Next, check if torch.distributed is initialized and has them set; yes? use them.
        # 4. If none of those are available, rank=0 and world_size=1.
        if "WORLD_SIZE" in os.environ and "RANK" in os.environ:
            # If deepspeed launcher is being used, it will set the env variables automatically.
            self.world_size = ifnone(world_size, int(os.environ["WORLD_SIZE"]))
            self.rank = ifnone(rank, int(os.environ["RANK"]))
        elif dist.is_available() and dist.is_initialized():
            self.world_size = ifnone(world_size, dist.get_world_size())
            self.rank = ifnone(rank, dist.get_rank())
        else:
            self.world_size = ifnone(world_size, 1)
            self.rank = ifnone(rank, 0)
        assert self.rank < self.world_size

    def set_epoch(self, epoch: int) -> None:
        """
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        :param epoch: Epoch number.
        """
        if self.epoch != epoch:
            # Changing the epoch automatically tells the sampler to discard the progress
            # from a previously read state dict.
            self.allow_iter_to_reset_state()
        self.epoch = epoch
        self.diagnostics.set_epoch(epoch)

    def filter(self, predicate: Callable[[Cut], bool]) -> "CutSampler":
        """
        Add a constraint on individual cuts that has to be satisfied to consider them.

        Can be useful when handling large, lazy manifests where it is not feasible to
        pre-filter them before instantiating the sampler.

        Example:
            >>> cuts = CutSet(...)
            ... sampler = SimpleCutSampler(cuts, max_duration=100.0)
            ... # Retain only the cuts that have at least 1s and at most 20s duration.
            ... sampler.filter(lambda cut: 1.0 <= cut.duration <= 20.0)
        """
        if isinstance(self._filter_fn, _filter_nothing):
            self._filter_fn = predicate
        else:
            self._filter_fn = _and(self._filter_fn, predicate)
        return self

    def map(self, fn: Callable[[CutSet], CutSet]) -> "CutSampler":
        """Apply ``fn`` to each mini-batch of ``CutSet`` before yielding it."""
        assert callable(
            fn
        ), f"Expected a callable accepting and returning a CutSet, received: '{fn}'"
        self._transforms.append(fn)
        return self

    def state_dict(self) -> Dict[str, Any]:
        """
        Return the current state of the sampler in a state_dict.
        Together with ``load_state_dict()``, this can be used to restore the
        training loop's state to the one stored in the state_dict.
        """
        return {
            "epoch": self.epoch,
            "drop_last": self.drop_last,
            "world_size": self.world_size,
            "rank": self.rank,
            "seed": self.seed,
            "shuffle": self.shuffle,
            "diagnostics": self.diagnostics.state_dict(),
        }

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Restore the state of the sampler that is described in a state_dict.
        This will result in the sampler yielding batches from where the previous training left it off.

        .. caution::
            The samplers are expected to be initialized with the same CutSets,
            but this is not explicitly checked anywhere.

        .. caution::
            The input ``state_dict`` is being mutated: we remove each consumed key, and expect
            it to be empty at the end of loading. If you don't want this behavior, pass a copy
            inside of this function (e.g., using ``import deepcopy``).

        .. note::
            For implementers of sub-classes of CutSampler: the flag ``self._just_restored_state`` has to be
            handled in ``__iter__`` to make it avoid resetting the just-restored state (only once).
        """
        self.drop_last = state_dict.pop("drop_last")
        world_size = state_dict.pop("world_size")
        assert self.world_size == world_size, (
            f"Cannot restore sampler with a different world_size (before load_state_dict(): {self.world_size},"
            f"attempted restoring to {world_size}). Changing the world_size would result in different batches "
            f"being returned from the sampler."
        )
        # We are explicitly discarding the "rank" argument to support restoring multi-GPU training
        # without too much hassle.
        # We assume that if the world_size is OK, the samplers local ranks are fine.
        del state_dict["rank"]
        assert self.seed == state_dict.pop("seed")
        shuffle = state_dict.pop("shuffle")
        if self.shuffle != shuffle:
            warnings.warn(
                "Overriding the shuffle value in CutSampler based on state_dict"
                f"(initialized to {self.shuffle}; restored to {shuffle})."
            )
        self.shuffle = shuffle
        self.epoch = state_dict.pop("epoch")
        self.diagnostics.load_state_dict(state_dict.pop("diagnostics"))
        assert (
            len(state_dict) == 0
        ), "Error in CutSampler.load_state_dict(): Unexpected keys:\n- " + "\n- ".join(
            state_dict.keys()
        )
        self._just_restored_state = True

    def __iter__(self):
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement __iter__()"
        )

    def _next_batch(self):
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement self._next_batch()"
        )

    @property
    def remaining_duration(self) -> Optional[float]:
        """
        Remaining duration of data left in the sampler (may be inexact due to float arithmetic).
        Not available when the CutSet is read in lazy mode (returns None).
        """
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement self.remaining_duration"
        )

    @property
    def remaining_cuts(self) -> Optional[int]:
        """
        Remaining number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).
        """
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement self.remaining_cuts"
        )

    @property
    def num_cuts(self) -> Optional[int]:
        """
        Total number of cuts in the sampler.
        Not available when the CutSet is read in lazy mode (returns None).
        """
        raise NotImplementedError(
            "Sub-classes of CutSampler have to implement self.num_cuts"
        )

    def allow_iter_to_reset_state(self):
        """
        Enables re-setting to the start of an epoch when iter() is called.
        This is only needed in one specific scenario: when we restored previous
        sampler state via ``sampler.load_state_dict()`` but want to discard
        the progress in the current epoch and start from the beginning.
        """
        self._just_restored_state = False

    def __next__(self):
        self.allow_iter_to_reset_state()
        # We use the following trick to ensure equal number of batches for each distributed
        # worker:
        # Every time a next batch is required, we will sample self.world_size batches first,
        # and then return the one at position self.rank.
        # When world_size=1 (i.e., single device) this doesn't change anything.
        # When world_size>1 (i.e., multi-GPU) the behavior depends on the setting of ``drop_last``.
        # To prevent some rank from terminating the iteration earlier than others (and typically going into deadlock),
        # we will either:
        # a) [drop_last=False, default] redistribute the examples to yield a (partial) mini-batch in each rank
        #       (if there's not enough examples, we'll duplicate some);
        # b) [drop_last=True] we'll stop early and discard the mini-batches for all ranks that had them.
        #       Note that drop_last=True implies the last partial mini-batch would also be discarded in a
        #       single-GPU setting, to be consistent with PyTorch's drop_last logic.
        batches = []
        for _ in range(self.world_size):
            try:
                batch = self._next_batch()
                batches.append(batch)
            except StopIteration:
                if self.world_size == 1 or self.drop_last:
                    # The users indicated they want an equal number of batches on all
                    # ranks and are ready to lose some data: drop remainder batches.
                    raise
                # If we got here, it means there's one or more empty mini-batch that
                # hasn't triggered StopIteration. This scenario is handled below.

        if len(batches) == 0:
            raise StopIteration()  # normal end of iteration when drop_last=False
        elif len(batches) != self.world_size:
            # "From each according to his ability, to each according to his needs."
            # We hit the end of data and at least one rank is left without a mini-batch.
            # Since we have access to the mini-batches in all ranks here, we can
            # deterministically re-distribute the examples to yield a partial mini-batch
            # in every rank (i.e., the result of the redistribution doesn't depend on rank).
            # The only problematic scenario is when the number of examples is smaller
            # than world size. In these cases, we will duplicate the first ``n_diff``
            # examples so that each rank has exactly 1 example in its mini-batch.
            combined = combine([b for b in batches if b is not None])
            chunk = 0
            while (diff := self.world_size - len(combined)) > 0:
                combined = combined + combined.subset(first=diff).modify_ids(
                    mark_as_duplicate(chunk)
                )
                chunk += 1
            batches = combined.split(self.world_size)

        selected = batches[self.rank]
        self._log_diagnostics(selected)
        for tfn in self._transforms:
            selected = tfn(selected)
        attach_dataloading_info(selected, rank=self.rank, world_size=self.world_size)
        return selected

    def _log_diagnostics(self, batch: Union[CutSet, Tuple[CutSet, ...]]) -> None:
        if isinstance(batch, CutSet):
            self.diagnostics.keep(batch)
        elif isinstance(batch, tuple) and isinstance(batch[0], CutSet):
            self.diagnostics.keep(batch[0])
        else:
            raise ValueError(f"Object with unexpected type: {batch}")

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        return self.diagnostics.get_report()


def mark_as_duplicate(iteration: int) -> Callable[[str], str]:
    def inner(cut_id: str) -> str:
        return f"{cut_id}_dup{iteration}"

    return inner


def attach_dataloading_info(cuts: CutSet, rank: int, world_size: int) -> None:
    """
    Attaches diagnostic info about dataloading to each cut under ``dataloading_info`` custom field.
    This information contains the rank, world_size, and worker_id.
    If the training is not distributed, rank and world_size are 0 and 1.
    If the num_workers argument in DataLoader was 0, worker_id is None.
    """
    worker_info = torch.utils.data.get_worker_info()
    if worker_info is None:
        worker_id = None
    else:
        worker_id = worker_info.id
    info = {"rank": rank, "world_size": world_size, "worker_id": worker_id}
    for cut in cuts:
        cut.dataloading_info = info


class SamplingConstraint(metaclass=ABCMeta):
    """
    Defines the interface for sampling constraints. A sampling constraint
    keeps track of the sampled examples and lets the sampler know when it
    should yield a mini-batch.
    """

    @abstractmethod
    def add(self, example: Any) -> None:
        """
        Update the sampling constraint with the information about the sampled example
        (e.g. current batch size, total duration).
        """
        pass

    @abstractmethod
    def exceeded(self) -> bool:
        """Inform if the sampling constraint has been exceeded."""
        pass

    @abstractmethod
    def close_to_exceeding(self) -> bool:
        """Inform if we're going to exceed the sampling constraint after adding one more example."""
        pass

    @abstractmethod
    def reset(self) -> None:
        """Resets the internal state (called after yielding a mini-batch)."""
        pass

    @abstractmethod
    def measure_length(self, example: Any) -> float:
        """
        Returns the "size" of an example, used to create bucket distribution for bucketing samplers
        (e.g., for audio it may be duration; for text it may be number of tokens; etc.).
        """
        pass

    def copy(self) -> "SamplingConstraint":
        """Return a shallow copy of this constraint."""
        return copy.copy(self)


@dataclass
class TimeConstraint(SamplingConstraint):
    """
    Represents a time-based constraint for sampler classes.
    It is defined as maximum total batch duration (in seconds) and/or the total number of cuts.

    :class:`TimeConstraint` can be used for tracking whether the criterion has been exceeded
    via the `add(cut)`, `exceeded()` and `reset()` methods.
    It will automatically track the right criterion (i.e. select duration from the cut).
    It can also be a null constraint (never exceeded).

    When ``quadratic_duration`` is set, we will try to compensate for models that have a
    quadratic complexity w.r.t. the input sequence length. We use the following formula
    to determine the effective duration for each cut::

        effective_duration = duration + (duration ** 2) / quadratic_duration

    We recomend setting quadratic_duration to something between 15 and 40 for transformer architectures.
    """

    max_duration: Optional[Seconds] = None
    max_cuts: Optional[int] = None
    current: Union[int, Seconds] = 0
    num_cuts: int = 0
    longest_seen: Union[int, float] = 0
    quadratic_duration: Optional[Seconds] = None

    def __post_init__(self) -> None:
        assert is_none_or_gt(self.max_duration, 0)
        assert is_none_or_gt(self.max_cuts, 0)
        assert is_none_or_gt(self.quadratic_duration, 0)

    def is_active(self) -> bool:
        """Is it an actual constraint, or a dummy one (i.e. never exceeded)."""
        return self.max_duration is not None or self.max_cuts is not None

    def add(self, example: Cut) -> None:
        """
        Increment the internal counter for the time constraint,
        selecting the right property from the input ``cut`` object.
        """
        if self.max_duration is not None:
            duration = self._maybe_apply_quadratic_correction(example.duration)
            self.current += duration
            self.longest_seen = max(self.longest_seen, duration)
        self.num_cuts += 1

    def _maybe_apply_quadratic_correction(self, duration: Seconds) -> Seconds:
        if self.quadratic_duration is None:
            return duration
        # For the quadratic complexity case, we add a term that accounts for
        # extra memory occupied by the model. The 1/quadratic_duration term causes
        # the effective duration to be doubled when it's equal to quadratic_duration.
        return duration + (duration**2) / self.quadratic_duration

    def exceeded(self) -> bool:
        """Is the constraint exceeded or not."""
        if self.max_cuts is not None and self.num_cuts > self.max_cuts:
            return True
        if self.max_duration is None:
            return False
        effective_duration = self.num_cuts * self.longest_seen
        return effective_duration > self.max_duration

    def close_to_exceeding(self) -> bool:
        """
        Check if the batch is close to satisfying the constraints.
        We define "closeness" as: if we added one more cut that has
        duration/num_frames/num_samples equal to the longest seen cut
        in the current batch, then the batch would have exceeded the constraints.
        """
        if self.max_cuts is not None and self.num_cuts >= self.max_cuts:
            return True

        if self.max_duration is not None:
            effective_duration = (self.num_cuts + 1) * self.longest_seen
            return effective_duration > self.max_duration
        return False

    def reset(self) -> None:
        """
        Reset the internal counter (to be used after a batch was created,
        to start collecting a new one).
        """
        self.current = 0
        self.num_cuts = 0
        self.longest_seen = 0

    def measure_length(self, example: Cut) -> float:
        return example.duration

    def state_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        self.max_duration = state_dict.pop("max_duration")
        self.max_cuts = state_dict.pop("max_cuts")
        self.current = state_dict.pop("current")
        self.num_cuts = state_dict.pop("num_cuts")
        self.longest_seen = state_dict.pop("longest_seen", 0)
        self.quadratic_duration = state_dict.pop("quadratic_duration", None)
        # backward compatibility
        state_dict.pop("strict", None)
        state_dict.pop("max_samples", None)
        state_dict.pop("max_frames", None)
        assert len(state_dict) == 0, (
            "Error in TimeConstraint.load_state_dict(): Unexpected keys:\n- "
            + "\n- ".join(state_dict.keys())
        )

    def __add__(self, other: "TimeConstraint") -> "TimeConstraint":
        for key in ("max_duration", "max_cuts", "quadratic_duration"):
            self_attr = getattr(self, key)
            other_attr = getattr(other, key)
            is_none = self_attr is None and other_attr is None
            assert is_none or isclose(self_attr, other_attr), (
                f"To add two TimeConstraint objects, they need to represent the same constraint "
                f"(got self.{key}={self_attr} != other.{key}={other_attr})."
            )
        return TimeConstraint(
            max_duration=self.max_duration,
            max_cuts=self.max_cuts,
            current=self.current + other.current,
            num_cuts=self.num_cuts + other.num_cuts,
            longest_seen=max(self.longest_seen, other.longest_seen),
            quadratic_duration=self.quadratic_duration,
        )

    def __eq__(self, other: "TimeConstraint") -> bool:
        return (
            self.max_duration == other.max_duration
            and self.max_cuts == other.max_cuts
            and self.quadratic_duration == other.quadratic_duration
        )


@dataclass
class TokenConstraint(SamplingConstraint):
    """
    Represents a token-based constraint for sampler classes that sample text data.
    It is defined as maximum total number of tokens in a mini-batch and/or max batch size.

    Similarly to :class:`TimeConstraint`, we support ``quadratic_length`` for quadratic
    token penalty when sampling longer texts.
    """

    max_tokens: int = None
    max_examples: int = None
    current: int = 0
    num_examples: int = 0
    longest_seen: int = 0
    quadratic_length: Optional[int] = None

    def __post_init__(self) -> None:
        assert is_none_or_gt(self.max_tokens, 0)
        assert is_none_or_gt(self.max_examples, 0)
        assert is_none_or_gt(self.quadratic_length, 0)

    def add(self, example: TextExample) -> None:
        """
        Increment the internal token counter for the constraint,
        selecting the right property from the input object.
        """
        if self.max_tokens is not None:
            size = self._maybe_apply_quadratic_correction(self.measure_length(example))
            self.current += size
            self.longest_seen = max(self.longest_seen, size)
        self.num_examples += 1

    def _maybe_apply_quadratic_correction(self, size: int) -> int:
        if self.quadratic_length is None:
            return size
        # For the quadratic complexity case, we add a term that accounts for
        # extra memory occupied by the model. The 1/quadratic_length term causes
        # the effective length to be doubled when it's equal to quadratic_length.
        return size + (size**2) / self.quadratic_length

    def exceeded(self) -> bool:
        """Is the constraint exceeded or not."""
        if self.max_examples is not None and self.num_examples > self.max_examples:
            return True
        if self.max_tokens is None:
            return False
        effective_duration = self.num_examples * self.longest_seen
        return effective_duration > self.max_tokens

    def close_to_exceeding(self) -> bool:
        """
        Check if the batch is close to satisfying the constraints.
        We define "closeness" as: if we added one more cut that has
        duration/num_frames/num_samples equal to the longest seen cut
        in the current batch, then the batch would have exceeded the constraints.
        """
        if self.max_examples is not None and self.num_examples >= self.max_examples:
            return True

        if self.max_tokens is not None:
            effective_size = (self.num_examples + 1) * self.longest_seen
            return effective_size > self.max_tokens
        return False

    def reset(self) -> None:
        """
        Reset the internal counter (to be used after a batch was created,
        to start collecting a new one).
        """
        self.current = 0
        self.num_examples = 0
        self.longest_seen = 0

    def measure_length(self, example: TextExample) -> float:
        return example.num_tokens


@dataclass
class EpochDiagnostics:
    epoch: int = 0
    kept_cuts: int = 0
    discarded_cuts: int = 0
    kept_batches: int = 0
    discarded_batches: int = 0

    @property
    def total_cuts(self) -> int:
        return self.kept_cuts + self.discarded_cuts

    @property
    def total_batches(self) -> int:
        return self.kept_batches + self.discarded_batches

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        if self.total_batches == 0 or self.total_cuts == 0:
            return (
                "Sampling statistics unavailable: EpochDiagnostics received no cuts or batches. "
                "If this is unexpected, and you're using a custom sampler, ensure that the sampler "
                "is registering the batches in SamplerDiagnostics/EpochDiagnostics."
            )

        return (
            f"| ep {self.epoch:>3d} | cuts kept {self.kept_cuts:d}/{self.total_cuts:d} "
            f"({self.kept_cuts / self.total_cuts:.2%}) "
            f"| cuts discarded {self.discarded_cuts:d} "
            f"| batches kept {self.kept_batches:d}/{self.total_batches:d} "
            f"({self.kept_batches / self.total_batches:.2%})"
            f"| batches discarded {self.discarded_batches:d} |"
        )

    def state_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def load_state_dict(self, state_dict: Dict[str, Any]) -> "EpochDiagnostics":
        self.epoch = state_dict.pop("epoch")
        self.kept_batches = state_dict.pop("kept_batches")
        self.discarded_batches = state_dict.pop("discarded_batches")
        self.kept_cuts = state_dict.pop("kept_cuts")
        self.discarded_cuts = state_dict.pop("discarded_cuts")

        assert len(state_dict) == 0, (
            "Error in EpochDiagnostics.load_state_dict(): Unexpected keys:\n- "
            + "\n- ".join(state_dict.keys())
        )

        return self

    def __add__(self, other: "EpochDiagnostics") -> "EpochDiagnostics":
        assert self.epoch == other.epoch
        return EpochDiagnostics(
            epoch=self.epoch,
            kept_cuts=self.kept_cuts + other.kept_cuts,
            kept_batches=self.kept_batches + other.kept_batches,
            discarded_cuts=self.discarded_cuts + other.discarded_cuts,
            discarded_batches=self.discarded_batches + other.discarded_batches,
        )


@dataclass
class SamplingDiagnostics:
    """
    Utility for collecting diagnostics about the sampling process:
    how many cuts/batches were discarded.
    """

    current_epoch: int = 0
    stats_per_epoch: Dict[int, EpochDiagnostics] = None

    def __post_init__(self):
        if self.stats_per_epoch is None:
            self.stats_per_epoch = {}
            self.set_epoch(self.current_epoch)

    def reset_current_epoch(self) -> None:
        self.stats_per_epoch[self.current_epoch] = EpochDiagnostics(self.current_epoch)

    def set_epoch(self, epoch: int) -> None:
        self.current_epoch = epoch
        if epoch not in self.stats_per_epoch:
            self.stats_per_epoch[epoch] = EpochDiagnostics(epoch=epoch)

    def advance_epoch(self) -> None:
        self.set_epoch(self.current_epoch + 1)

    @property
    def current_epoch_stats(self) -> EpochDiagnostics:
        return self.stats_per_epoch[self.current_epoch]

    def keep(self, cuts: Iterable[Cut]) -> None:
        cntr = 0
        for cut in cuts:
            self.current_epoch_stats.kept_cuts += 1
            cntr += 1
        if not cntr:
            warnings.warn(
                "Found and accepted batch with zero cuts. This could be an error."
            )
        self.current_epoch_stats.kept_batches += 1

    def discard(self, cuts: Iterable[Cut]) -> None:
        cntr = 0
        for cut in cuts:
            self.current_epoch_stats.discarded_cuts += 1
            cntr += 1
        if cntr:
            # We don't warn about discarded batches with 0 cuts.
            self.current_epoch_stats.discarded_batches += 1

    def discard_single(self, cut: Cut) -> None:
        self.current_epoch_stats.discarded_cuts += 1

    @property
    def kept_cuts(self) -> int:
        return sum(s.kept_cuts for s in self.stats_per_epoch.values())

    @property
    def discarded_cuts(self) -> int:
        return sum(s.discarded_cuts for s in self.stats_per_epoch.values())

    @property
    def kept_batches(self) -> int:
        return sum(s.kept_batches for s in self.stats_per_epoch.values())

    @property
    def discarded_batches(self) -> int:
        return sum(s.discarded_batches for s in self.stats_per_epoch.values())

    @property
    def total_cuts(self) -> int:
        return sum(s.total_cuts for s in self.stats_per_epoch.values())

    @property
    def total_batches(self) -> int:
        return sum(s.total_batches for s in self.stats_per_epoch.values())

    def get_report(self, per_epoch: bool = False) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        if self.total_batches == 0 or self.total_cuts == 0:
            return (
                "Sampling statistics unavailable: the SamplerDiagnostics received no cuts or batches. "
                "If this is unexpected, and you're using a custom sampler, ensure that the sampler "
                "is registering the batches in SamplerDiagnostics."
            )

        ret = []

        if per_epoch:
            for epoch in sorted(self.stats_per_epoch):
                ret.append(self.stats_per_epoch[epoch].get_report())

        ret.append(
            f"|  total  | cuts kept {self.kept_cuts:d}/{self.total_cuts:d} "
            f"({self.kept_cuts / self.total_cuts:.2%}) "
            f"| cuts discarded {self.discarded_cuts:d} "
            f"| batches kept {self.kept_batches:d}/{self.total_batches:d} "
            f"({self.kept_batches / self.total_batches:.2%})"
            f"| batches discarded {self.discarded_batches:d} |"
        )

        return "\n".join(ret)

    def state_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def load_state_dict(self, state_dict: Dict[str, Any]) -> "SamplingDiagnostics":
        self.current_epoch = state_dict.pop("current_epoch")
        self.stats_per_epoch = {
            epoch: EpochDiagnostics().load_state_dict(sd)
            for epoch, sd in state_dict.pop("stats_per_epoch").items()
        }
        return self

    def __add__(self, other: "SamplingDiagnostics") -> "SamplingDiagnostics":
        stats_per_epoch = deepcopy(self.stats_per_epoch)
        for epoch, stats in other.stats_per_epoch.items():
            if epoch in stats_per_epoch:
                stats_per_epoch[epoch] = stats_per_epoch[epoch] + stats
            else:
                stats_per_epoch[epoch] = stats

        return SamplingDiagnostics(
            current_epoch=self.current_epoch, stats_per_epoch=stats_per_epoch
        )


class _filter_nothing:
    def __call__(self, cut: Cut) -> bool:
        return True


def _and(
    fn1: Callable[[Cut], bool], fn2: Callable[[Cut], bool]
) -> Callable[[Cut], bool]:
    def _and_wrapper(cut: Cut) -> bool:
        return fn1(cut) and fn2(cut)

    return _and_wrapper



import logging
import os
import re
import sys
import traceback
from contextlib import contextmanager
from functools import lru_cache
from io import BytesIO, IOBase
from pathlib import Path
from subprocess import PIPE, CalledProcessError, run
from typing import Any, Generator, List, NamedTuple, Optional, Tuple, Union

import numpy as np
import torch

from lhotse.audio.utils import (
    AudioLoadingError,
    AudioSavingError,
    VideoInfo,
    verbose_audio_loading_exceptions,
)
from lhotse.augmentation import Resample
from lhotse.utils import Pathlike, Seconds, compute_num_samples, is_torchaudio_available

_FFMPEG_TORCHAUDIO_INFO_ENABLED: bool = is_torchaudio_available()
CURRENT_AUDIO_BACKEND: Optional["AudioBackend"] = None


def available_audio_backends() -> List[str]:
    """
    Return a list of names of available audio backends, including "default".
    """
    return ["default"] + sorted(AudioBackend.KNOWN_BACKENDS.keys())


@contextmanager
def audio_backend(
    backend: Union["AudioBackend", str]
) -> Generator["AudioBackend", None, None]:
    """
    Context manager that sets Lhotse's audio backend to the specified value
    and restores the previous audio backend at the end of its scope.

    Example::

        >>> with audio_backend("LibsndfileBackend"):
        ...     some_audio_loading_fn()
    """
    previous = get_current_audio_backend()
    b = set_current_audio_backend(backend)
    yield b
    set_current_audio_backend(previous)


def get_current_audio_backend() -> "AudioBackend":
    """
    Return the audio backend currently set by the user, or default.
    """
    global CURRENT_AUDIO_BACKEND

    # First check if the user has programmatically overridden the audio backend.
    if CURRENT_AUDIO_BACKEND is not None:
        return CURRENT_AUDIO_BACKEND

    # Then, check if the user has overridden the audio backend via an env var.
    maybe_backend = os.environ.get("LHOTSE_AUDIO_BACKEND")
    if maybe_backend is not None:
        set_current_audio_backend(maybe_backend)
        return CURRENT_AUDIO_BACKEND

    # Lastly, fall back to the default backend.
    set_current_audio_backend("default")
    return CURRENT_AUDIO_BACKEND


def set_current_audio_backend(backend: Union["AudioBackend", str]) -> "AudioBackend":
    """
    Force Lhotse to use a specific audio backend to read every audio file,
    overriding the default behaviour of educated guessing + trial-and-error.

    Example forcing Lhotse to use ``audioread`` library for every audio loading operation::

        >>> set_current_audio_backend(AudioreadBackend())
    """
    global CURRENT_AUDIO_BACKEND
    if backend == "default":
        backend = get_default_audio_backend()
    elif isinstance(backend, str):
        backend = AudioBackend.new(backend)
    else:
        if isinstance(backend, type):
            backend = backend()
        assert isinstance(
            backend, AudioBackend
        ), f"Expected str or AudioBackend, got: {backend}"
    CURRENT_AUDIO_BACKEND = backend
    return CURRENT_AUDIO_BACKEND


@lru_cache(maxsize=1)
def get_default_audio_backend() -> "AudioBackend":
    """
    Return a backend that can be used to read all audio formats supported by Lhotse.

    It first looks for special cases that need very specific handling
    (such as: opus, sphere/shorten, in-memory buffers)
    and tries to match them against relevant audio backends.

    Then, it tries to use several audio loading libraries (torchaudio, soundfile, audioread).
    In case the first fails, it tries the next one, and so on.
    """
    backends = []
    if os.environ.get("LHOTSE_LEGACY_OPUS_LOADING", "0") == "1":
        # We no longer use subprocess ffmpeg for OPUS, preferring libnsdfile backend instead.
        # However, in some cases users might have manifests created with the legacy backend
        # so we add this as an option to support them.
        backends.append(FfmpegSubprocessOpusBackend())
    backends += [
        # Use sph2pipe for .sph and shorten encoded audio
        Sph2pipeSubprocessBackend(),
        # Libsndfile seems to be the most stable backend in terms of covered formats and performance.
        LibsndfileBackend(),
        # New FFMPEG backend available only in torchaudio 2.0.x+
        TorchaudioFFMPEGBackend(),
        # Torchaudio should be able to deal with most audio types...
        TorchaudioDefaultBackend(),
        # ... if not, try audioread as a hail mary...
        AudioreadBackend(),
        # ... oops.
    ]
    return CompositeAudioBackend(backends)


def set_ffmpeg_torchaudio_info_enabled(enabled: bool) -> None:
    """
    Override Lhotse's global setting for whether to use ffmpeg-torchaudio to
    compute the duration of audio files. If disabled, we fall back to using a different
    backend such as sox_io or soundfile.

    .. note:: See this issue for more details: https://github.com/lhotse-speech/lhotse/issues/1026

    Example::

        >>> import lhotse
        >>> lhotse.set_ffmpeg_torchaudio_info_enabled(False)  # don't use ffmpeg-torchaudio

    :param enabled: Whether to use torchaudio to compute audio file duration.
    """
    global _FFMPEG_TORCHAUDIO_INFO_ENABLED
    if enabled != _FFMPEG_TORCHAUDIO_INFO_ENABLED:
        logging.info(
            "The user overrided the global setting for whether to use ffmpeg-torchaudio "
            "to compute the duration of audio files. "
            f"Old setting: {_FFMPEG_TORCHAUDIO_INFO_ENABLED}. "
            f"New setting: {enabled}."
        )
    _FFMPEG_TORCHAUDIO_INFO_ENABLED = enabled


def get_ffmpeg_torchaudio_info_enabled() -> bool:
    """
    Return FFMPEG_TORCHAUDIO_INFO_ENABLED, which is Lhotse's global setting for whether to
    use ffmpeg-torchaudio to compute the duration of audio files.

    Example::

        >>> import lhotse
        >>> lhotse.get_ffmpeg_torchaudio_info_enabled()
    """
    return _FFMPEG_TORCHAUDIO_INFO_ENABLED


FileObject = Any  # Alias for file-like objects


class AudioBackend:
    """
    Internal Lhotse abstraction. An AudioBackend defines three methods:
    one for reading audio, and two filters that help determine if it should be used.

    ``handles_special_case`` means this backend should be exclusively
    used for a given type of input path/file.

    ``is_applicable`` means this backend most likely can be used for a given type of input path/file,
    but it may also fail. Its purpose is more to filter out formats that definitely are not supported.
    """

    KNOWN_BACKENDS = {}

    def __init_subclass__(cls, **kwargs):
        if cls.__name__ not in AudioBackend.KNOWN_BACKENDS:
            AudioBackend.KNOWN_BACKENDS[cls.__name__] = cls
        super().__init_subclass__(**kwargs)

    @classmethod
    def new(cls, name: str) -> "AudioBackend":
        if name not in cls.KNOWN_BACKENDS:
            raise RuntimeError(f"Unknown audio backend name: {name}")
        return cls.KNOWN_BACKENDS[name]()

    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        raise NotImplementedError()

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        raise NotImplementedError()

    def handles_special_case(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return False

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return True

    def supports_save(self) -> bool:
        return False

    def supports_info(self) -> bool:
        return False

    def save_audio(
        self,
        dest: Union[str, Path, BytesIO],
        src: Union[torch.Tensor, np.ndarray],
        sampling_rate: int,
        format: Optional[str] = None,
        encoding: Optional[str] = None,
    ) -> None:
        raise NotImplementedError()


class FfmpegSubprocessOpusBackend(AudioBackend):
    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ):
        assert isinstance(
            path_or_fd, (str, Path)
        ), f"Cannot use an ffmpeg subprocess to read from path of type: '{type(path_or_fd)}'"
        return read_opus_ffmpeg(
            path=path_or_fd,
            offset=offset,
            duration=duration,
            force_opus_sampling_rate=force_opus_sampling_rate,
        )

    def handles_special_case(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return isinstance(path_or_fd, (str, Path)) and str(path_or_fd).lower().endswith(
            ".opus"
        )

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return self.handles_special_case(path_or_fd)

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return opus_info(path_or_fd, force_opus_sampling_rate)


class Sph2pipeSubprocessBackend(AudioBackend):
    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        assert isinstance(
            path_or_fd, (str, Path)
        ), f"Cannot use an sph2pipe subprocess to read from path of type: '{type(path_or_fd)}'"
        return read_sph(
            sph_path=path_or_fd,
            offset=offset,
            duration=duration,
        )

    def handles_special_case(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return isinstance(path_or_fd, (str, Path)) and str(path_or_fd).lower().endswith(
            ".sph"
        )

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return self.handles_special_case(path_or_fd)

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return sph_info(path_or_fd)


class FfmpegTorchaudioStreamerBackend(AudioBackend):
    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        return torchaudio_ffmpeg_load(
            path_or_fileobj=path_or_fd,
            offset=offset,
            duration=duration,
        )

    def handles_special_case(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return (
            is_torchaudio_available()
            and torchaudio_supports_ffmpeg()
            and isinstance(path_or_fd, BytesIO)
        )

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        # Technically it's applicable with regular files as well, but for now
        # we're not enabling that feature.
        return (
            is_torchaudio_available()
            and torchaudio_supports_ffmpeg()
            and isinstance(path_or_fd, BytesIO)
        )

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return torchaudio_info(path_or_fd)


class TorchaudioDefaultBackend(AudioBackend):
    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        return torchaudio_load(
            path_or_fd=path_or_fd,
            offset=offset,
            duration=duration,
        )

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return is_torchaudio_available()

    def supports_save(self) -> bool:
        return True

    def save_audio(
        self,
        dest: Union[str, Path, BytesIO],
        src: Union[torch.Tensor, np.ndarray],
        sampling_rate: int,
        format: Optional[str] = None,
        encoding: Optional[str] = None,
    ) -> None:
        import torchaudio

        assert (
            torchaudio_soundfile_supports_format()
        ), "We don't support saving audio with torchaudio pre v0.9.0"
        src = torch.as_tensor(src)
        if isinstance(dest, Path):
            dest = str(dest)
        saving_flac = format.lower() == "flac" or (
            not isinstance(dest, BytesIO) and dest.endswith(".flac")
        )
        if saving_flac:
            # Prefer saving with soundfile backend whenever possible to avoid issue:
            # https://github.com/pytorch/audio/issues/2662
            # Saving with sox_io backend to FLAC may corrupt the file.
            torchaudio.backend.soundfile_backend.save(
                dest,
                src,
                sample_rate=sampling_rate,
                format="flac",
                bits_per_sample=16,
            )
        else:
            # Using whichever backend is enabled in torchaudio.
            torchaudio.save(
                dest, src, sample_rate=sampling_rate, format=format, encoding=encoding
            )

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return torchaudio_info(path_or_fd)


class TorchaudioFFMPEGBackend(AudioBackend):
    """
    A new FFMPEG backend available in torchaudio 2.0.
    It should be free from many issues of soundfile and sox_io backends.
    """

    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        return torchaudio_2_ffmpeg_load(
            path_or_fd=path_or_fd,
            offset=offset,
            duration=duration,
            resample_rate=force_opus_sampling_rate,
        )

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        """
        FFMPEG backend requires at least Torchaudio 2.0.
        For version == 2.0.x, we also need env var TORCHAUDIO_USE_BACKEND_DISPATCHER=1
        For version >= 2.1.x, this will already be the default.
        """
        return is_torchaudio_available() and torchaudio_ffmpeg_backend_available()

    def supports_save(self) -> bool:
        return True

    def save_audio(
        self,
        dest: Union[str, Path, BytesIO],
        src: Union[torch.Tensor, np.ndarray],
        sampling_rate: int,
        format: Optional[str] = None,
        encoding: Optional[str] = None,
    ) -> None:
        import torchaudio

        if not torch.is_tensor(src):
            src = torch.as_tensor(src)
        if isinstance(dest, Path):
            dest = str(dest)
        torchaudio.save(
            dest,
            src,
            sample_rate=sampling_rate,
            format=format,
            encoding=encoding,
            backend="ffmpeg",
        )

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return torchaudio_ffmpeg_streamer_info(path_or_fd)


class LibsndfileBackend(AudioBackend):
    """
    A backend that uses PySoundFile.

    .. note:: PySoundFile has issues on MacOS because of the way its CFFI bindings are implemented.
        For now, we disable it on this platform.
        See: https://github.com/bastibe/python-soundfile/issues/331
    """

    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        return soundfile_load(
            path_or_fd=path_or_fd,
            offset=offset,
            duration=duration,
        )

    def handles_special_case(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        if (
            isinstance(path_or_fd, BytesIO)
            and not torchaudio_ffmpeg_backend_available()
        ):
            return True  # prefer this to old torchaudio for file IO
        if isinstance(path_or_fd, (Path, str)) and str(path_or_fd).endswith(".opus"):
            return True  # use libnsdfile for OPUS
        return False

    def is_applicable(self, path_or_fd: Union[Pathlike, FileObject]) -> bool:
        return True

    def supports_save(self) -> bool:
        return True

    def save_audio(
        self,
        dest: Union[str, Path, BytesIO],
        src: Union[torch.Tensor, np.ndarray],
        sampling_rate: int,
        format: Optional[str] = None,
        encoding: Optional[str] = None,
    ) -> None:
        import soundfile as sf

        if torch.is_tensor(src):
            src = src.numpy()
        if isinstance(dest, Path):
            dest = str(dest)
        if format == "opus":
            format = "ogg"
            encoding = "opus"
        sf.write(
            file=dest,
            data=src.T,
            samplerate=sampling_rate,
            format=format,
            subtype=encoding,
        )

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return soundfile_info(path_or_fd)


class AudioreadBackend(AudioBackend):
    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        return audioread_load(
            path_or_file=path_or_fd,
            offset=offset,
            duration=duration,
        )

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        return audioread_info(path_or_fd)


class CompositeAudioBackend(AudioBackend):
    """
    Combines multiple audio backends.
    It will try each out sequentially, and back off to the next one in the list if the current one fails.
    It uses the special filter methods to prioritize special case backends,
    and skip backends that are not applicable.
    """

    def __init__(self, backends: List[AudioBackend]):
        self.backends = backends

    def read_audio(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        force_opus_sampling_rate: Optional[int] = None,
    ) -> Tuple[np.ndarray, int]:
        candidates = []
        for b in self.backends:
            if b.handles_special_case(path_or_fd):
                candidates.append(b)

        assert len(candidates) < 2, (
            f"CompositeAudioBackend has more than one sub-backend that "
            f"handles a given special case for input '{path_or_fd}'"
        )

        if len(candidates) == 1:
            try:
                return candidates[0].read_audio(
                    path_or_fd=path_or_fd,
                    offset=offset,
                    duration=duration,
                    force_opus_sampling_rate=force_opus_sampling_rate,
                )
            except Exception as e:
                raise AudioLoadingError(
                    f"Reading audio from '{path_or_fd}' failed. Details: {type(e)}: {str(e)}"
                )

        exceptions = []
        for b in self.backends:
            if b.is_applicable(path_or_fd):
                try:
                    return b.read_audio(
                        path_or_fd=path_or_fd,
                        offset=offset,
                        duration=duration,
                        force_opus_sampling_rate=force_opus_sampling_rate,
                    )
                except Exception as e:
                    msg = f"Exception #{len(exceptions)} ({type(b)}): "
                    if verbose_audio_loading_exceptions():
                        exceptions.append(f"{msg}{traceback.format_exc()}")
                    else:
                        exceptions.append(f"{msg}{type(e)}: {str(e)}")

        if not exceptions:
            raise AudioLoadingError(
                f"No applicable backend found for input: '{path_or_fd}'"
            )
        else:
            NL = "\n"
            maybe_info = (
                ""
                if verbose_audio_loading_exceptions()
                else "\nSet LHOTSE_AUDIO_LOADING_EXCEPTION_VERBOSE=1 environment variable for full stack traces."
            )
            raise AudioLoadingError(
                f"Reading audio from '{path_or_fd}' failed. Details:{NL}{NL.join(exceptions)}{maybe_info}"
            )

    def supports_save(self) -> bool:
        return any(b.supports_save() for b in self.backends)

    def save_audio(
        self,
        dest: Union[str, Path, BytesIO],
        src: Union[torch.Tensor, np.ndarray],
        sampling_rate: int,
        format: Optional[str] = None,
        encoding: Optional[str] = None,
    ) -> None:
        candidates = [b for b in self.backends if b.supports_save()]

        exceptions = []
        for b in candidates:
            if not b.is_applicable(dest):
                continue
            try:
                return b.save_audio(
                    dest=dest,
                    src=src,
                    sampling_rate=sampling_rate,
                    format=format,
                    encoding=encoding,
                )
            except Exception as e:
                msg = f"Exception #{len(exceptions)} ({type(b)}): "
                if verbose_audio_loading_exceptions():
                    exceptions.append(f"{msg}{traceback.format_exc()}")
                else:
                    exceptions.append(f"{msg}{type(e)}: {str(e)}")

        if not exceptions:
            raise AudioSavingError(f"No applicable backend found for saving audio.")
        else:
            NL = "\n"
            maybe_info = (
                ""
                if verbose_audio_loading_exceptions()
                else "\nSet LHOTSE_AUDIO_LOADING_EXCEPTION_VERBOSE=1 environment variable for full stack traces."
            )
            raise AudioSavingError(
                f"Saving audio failed. Details:{NL}{NL.join(exceptions)}{maybe_info}"
            )

    def supports_info(self) -> bool:
        return True

    def info(
        self,
        path_or_fd: Union[Pathlike, FileObject],
        force_opus_sampling_rate: Optional[int] = None,
    ):
        backends = [b for b in self.backends if b.supports_info()]
        candidates = []
        for b in backends:
            if b.handles_special_case(path_or_fd):
                candidates.append(b)

        assert len(candidates) < 2, (
            f"CompositeAudioBackend has more than one sub-backend that "
            f"handles a given special case for input '{path_or_fd}'"
        )

        if len(candidates) == 1:
            try:
                return candidates[0].info(path_or_fd=path_or_fd)
            except Exception as e:
                raise AudioLoadingError(
                    f"Fetching info about audio from '{path_or_fd}' failed. Details: {type(e)}: {str(e)}"
                )

        exceptions = []
        for b in backends:
            if b.is_applicable(path_or_fd):
                try:
                    return b.info(path_or_fd=path_or_fd)
                except Exception as e:
                    msg = f"Exception #{len(exceptions)} ({type(b)}): "
                    if verbose_audio_loading_exceptions():
                        exceptions.append(f"{msg}{traceback.format_exc()}")
                    else:
                        exceptions.append(f"{msg}{type(e)}: {str(e)}")

        if not exceptions:
            raise AudioLoadingError(
                f"No applicable backend found for input: '{path_or_fd}'"
            )
        else:
            NL = "\n"
            maybe_info = (
                ""
                if verbose_audio_loading_exceptions()
                else "\nSet LHOTSE_AUDIO_LOADING_EXCEPTION_VERBOSE=1 environment variable for full stack traces."
            )
            raise AudioLoadingError(
                f"Fetching info about audio from '{path_or_fd}' failed. Details:{NL}{NL.join(exceptions)}{maybe_info}"
            )


class LibsndfileCompatibleAudioInfo(NamedTuple):
    channels: int
    frames: int
    samplerate: int
    duration: float
    video: Optional[VideoInfo] = None


@lru_cache(maxsize=1)
def torchaudio_supports_ffmpeg() -> bool:
    """
    Returns ``True`` when torchaudio version is at least 0.12.0, which
    has support for FFMPEG streamer API.
    """
    # If user has disabled ffmpeg-torchaudio, we don't need to check the version.
    if not _FFMPEG_TORCHAUDIO_INFO_ENABLED:
        return False
    return check_torchaudio_version_gt("0.12.0")


@lru_cache(maxsize=1)
def torchaudio_ffmpeg_backend_available() -> bool:
    """
    Returns ``True`` when torchaudio.load supports "ffmpeg" backend.
    This requires either version 2.1.x+
    """
    return is_torchaudio_available() and check_torchaudio_version_gt("2.1.0")


@lru_cache(maxsize=1)
def torchaudio_soundfile_supports_format() -> bool:
    """
    Returns ``True`` when torchaudio version is at least 0.9.0, which
    has support for ``format`` keyword arg in ``torchaudio.save()``.
    """
    return check_torchaudio_version_gt("0.9.0")


def check_torchaudio_version_gt(version: str) -> bool:
    import torchaudio
    from packaging import version as _version

    return _version.parse(torchaudio.__version__) >= _version.parse(version)


def torchaudio_info(
    path_or_fileobj: Union[Path, str, BytesIO]
) -> LibsndfileCompatibleAudioInfo:
    """
    Return an audio info data structure that's a compatible subset of ``pysoundfile.info()``
    that we need to create a ``Recording`` manifest.
    """
    import torchaudio

    if torchaudio_ffmpeg_backend_available():
        # Torchaudio 2.1 with official "ffmpeg" backend should solve all the special cases below.
        info = torchaudio.info(path_or_fileobj, backend="ffmpeg")
        return LibsndfileCompatibleAudioInfo(
            channels=info.num_channels,
            frames=info.num_frames,
            samplerate=int(info.sample_rate),
            duration=info.num_frames / info.sample_rate,
        )

    is_mpeg = isinstance(path_or_fileobj, (str, Path)) and any(
        str(path_or_fileobj).endswith(ext) for ext in (".mp3", ".m4a")
    )
    is_fileobj = isinstance(path_or_fileobj, BytesIO)
    if (is_mpeg or is_fileobj) and torchaudio_supports_ffmpeg():
        # Torchaudio 0.12 has a new StreamReader API that uses ffmpeg.
        #
        # They dropped support for using sox bindings in torchaudio.info
        # for MP3 files and implicitly delegate the call to ffmpeg.
        # Unfortunately, they always return num_frames/num_samples = 0,
        # as explained here: https://github.com/pytorch/audio/issues/2524
        # We have to work around by streaming the MP3 and counting the number
        # of samples.
        #
        # Unfortunately torchaudio also has issues with reading from file objects
        # sometimes, which apparently we can work around by using StreamReader API.
        # See:
        # - https://github.com/pytorch/audio/issues/2524#issuecomment-1223901818
        # - https://github.com/pytorch/audio/issues/2662
        from torchaudio.io import StreamReader

        streamer = StreamReader(
            src=str(path_or_fileobj) if is_mpeg else path_or_fileobj
        )
        assert streamer.num_src_streams == 1, (
            "Lhotse doesn't support files with more than one source stream yet "
            "(not to be confused with multi-channel)."
        )
        info = streamer.get_src_stream_info(streamer.default_audio_stream)
        streamer.add_basic_audio_stream(
            frames_per_chunk=int(info.sample_rate),
        )
        tot_samples = 0
        for (chunk,) in streamer.stream():
            tot_samples += chunk.shape[0]
        return LibsndfileCompatibleAudioInfo(
            channels=info.num_channels,
            frames=tot_samples,
            samplerate=int(info.sample_rate),
            duration=tot_samples / info.sample_rate,
        )

    info = torchaudio.info(path_or_fileobj)
    return LibsndfileCompatibleAudioInfo(
        channels=info.num_channels,
        frames=info.num_frames,
        samplerate=int(info.sample_rate),
        duration=info.num_frames / info.sample_rate,
    )


def torchaudio_ffmpeg_streamer_info(
    path_or_fileobj: Union[Path, str, BytesIO]
) -> LibsndfileCompatibleAudioInfo:
    from torchaudio.io import StreamReader

    is_fileobj = not isinstance(path_or_fileobj, Path)
    is_mpeg = not is_fileobj and any(
        str(path_or_fileobj).endswith(ext) for ext in (".mp3", ".mp4", ".m4a")
    )
    if not is_fileobj:
        path_or_fileobj = str(path_or_fileobj)
    stream = StreamReader(path_or_fileobj)

    # Collect the information about available video and audio streams.
    num_streams = stream.num_src_streams
    audio_streams = {}
    video_streams = {}
    for stream_idx in range(num_streams):
        info = stream.get_src_stream_info(stream_idx)
        if info.media_type == "video":
            video_streams[stream_idx] = info
        elif info.media_type == "audio":
            audio_streams[stream_idx] = info
        else:
            raise RuntimeError(f"Unexpected media_type: {info}")

    assert (
        len(video_streams) < 2
    ), f"Lhotse currently does not support more than one video stream in a file (found {len(video_streams)})."
    assert len(audio_streams) < 2, (
        f"Lhotse currently does not support files with more than a single FFMPEG "
        f"audio stream yet (found {len(audio_streams)}). "
        f"Note that this is not the same as multi-channel which is generally supported."
    )

    meta = {}

    if video_streams:
        ((video_stream_idx, video_stream),) = list(video_streams.items())
        tot_frames = video_stream.num_frames

        if tot_frames == 0:  # num frames not available in header/metadata
            stream.add_basic_video_stream(
                round(video_stream.frame_rate), stream_index=video_stream_idx
            )
            for (chunk,) in stream.stream():
                tot_frames += chunk.shape[0]
            stream.remove_stream(0)

        meta["video"] = VideoInfo(
            fps=video_stream.frame_rate,
            height=video_stream.height,
            width=video_stream.width,
            num_frames=tot_frames,
        )

    if audio_streams:
        ((audio_stream_idx, audio_stream),) = list(audio_streams.items())
        stream.add_basic_audio_stream(
            frames_per_chunk=int(audio_stream.sample_rate),
            stream_index=audio_stream_idx,
        )

        def _try_read_num_samples():
            if is_mpeg or is_fileobj:
                # These cases often have insufficient or corrupted metadata, so we might need to scan
                # the full audio stream to learn the actual number of frames. If video is available,
                # we can quickly verify before performing the costly reading.
                video_info = meta.get("video", None)
                if video_info is not None:
                    audio_duration = audio_stream.num_frames / audio_stream.sample_rate
                    # for now 1ms tolerance
                    if abs(audio_duration - video_info.duration) < 1e-3:
                        return audio_stream.num_frames
                return 0
            else:
                return audio_stream.num_frames

        tot_samples = _try_read_num_samples()
        if tot_samples == 0:
            # There was a mismatch between video and audio duration in metadata,
            # we'll have to read the file to figure it out.
            for (chunk,) in stream.stream():
                tot_samples += chunk.shape[0]

        meta.update(
            channels=audio_stream.num_channels,
            frames=tot_samples,
            samplerate=int(audio_stream.sample_rate),
            duration=tot_samples / audio_stream.sample_rate,
        )

    return LibsndfileCompatibleAudioInfo(**meta)


def torchaudio_load(
    path_or_fd: Pathlike, offset: Seconds = 0, duration: Optional[Seconds] = None
) -> Tuple[np.ndarray, int]:
    import torchaudio

    if isinstance(path_or_fd, Path):
        path_or_fd = str(path_or_fd)

    # Need to grab the "info" about sampling rate before reading to compute
    # the number of samples provided in offset / num_frames.
    frame_offset = 0
    num_frames = -1
    if offset > 0 or duration is not None:
        audio_info = torchaudio_info(path_or_fd)
        if offset > 0:
            frame_offset = compute_num_samples(offset, audio_info.samplerate)
        if duration is not None:
            num_frames = compute_num_samples(duration, audio_info.samplerate)
    if isinstance(path_or_fd, IOBase):
        # Set seek pointer to the beginning of the file as torchaudio.info
        # might have left it at the end of the header
        path_or_fd.seek(0)
    audio, sampling_rate = torchaudio.load(
        path_or_fd,
        frame_offset=frame_offset,
        num_frames=num_frames,
    )
    return audio.numpy(), int(sampling_rate)


def torchaudio_2_ffmpeg_load(
    path_or_fd: Union[Pathlike, BytesIO],
    offset: Seconds = 0,
    duration: Optional[Seconds] = None,
    resample_rate: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    import torchaudio

    if isinstance(path_or_fd, Path):
        path_or_fd = str(path_or_fd)

    # Need to grab the "info" about sampling rate before reading to compute
    # the number of samples provided in offset / num_frames.
    frame_offset = 0
    num_frames = -1
    if offset > 0 or duration is not None:
        audio_info = torchaudio.info(path_or_fd, backend="ffmpeg")
        if offset > 0:
            frame_offset = compute_num_samples(offset, audio_info.sample_rate)
        if duration is not None:
            num_frames = compute_num_samples(duration, audio_info.sample_rate)
    if isinstance(path_or_fd, IOBase):
        # Set seek pointer to the beginning of the file as torchaudio.info
        # might have left it at the end of the header
        path_or_fd.seek(0)
    audio, sampling_rate = torchaudio.load(
        path_or_fd,
        frame_offset=frame_offset,
        num_frames=num_frames,
        backend="ffmpeg",
    )
    return audio.numpy(), int(sampling_rate)


def torchaudio_ffmpeg_load(
    path_or_fileobj: Union[Path, str, BytesIO],
    offset: Seconds = 0,
    duration: Optional[Seconds] = None,
) -> Tuple[np.ndarray, int]:
    import torchaudio

    if not torchaudio_supports_ffmpeg():
        raise RuntimeError(
            "Using FFMPEG streamer backend for reading is supported only "
            "with PyTorch 1.12+ and torchaudio 0.12+"
        )

    if isinstance(path_or_fileobj, Path):
        path_or_fileobj = str(path_or_fileobj)

    streamer = torchaudio.io.StreamReader(src=path_or_fileobj)
    assert streamer.num_src_streams == 1, (
        "Lhotse doesn't support files with more than one source stream yet "
        "(not to be confused with multi-channel)."
    )
    info = streamer.get_src_stream_info(streamer.default_audio_stream)
    sampling_rate = int(info.sample_rate)

    if duration is not None:
        # Try to read whole audio in a single chunk.
        streamer.add_basic_audio_stream(
            frames_per_chunk=compute_num_samples(duration, sampling_rate)
        )
        streamer.seek(offset)
        (audio,) = next(streamer.stream())
        audio = audio.transpose(0, 1)
    else:
        # Read in 1 second chunks and concatenate (we don't know how much audio is incoming)
        streamer.add_basic_audio_stream(frames_per_chunk=sampling_rate)
        streamer.seek(offset)
        audio = torch.cat([t.transpose(0, 1) for t, in streamer.stream()], dim=0)

    # Return shape (num_channels, num_samples)
    return audio.numpy(), sampling_rate


def soundfile_load(
    path_or_fd: Pathlike, offset: Seconds = 0, duration: Optional[Seconds] = None
) -> Tuple[np.ndarray, int]:
    import soundfile as sf

    with sf.SoundFile(path_or_fd) as sf_desc:
        sampling_rate = sf_desc.samplerate
        if offset > 0:
            # Seek to the start of the target read
            sf_desc.seek(compute_num_samples(offset, sampling_rate))
        if duration is not None:
            frame_duration = compute_num_samples(duration, sampling_rate)
        else:
            frame_duration = -1
        # Load the target number of frames, and transpose to match librosa form
        return (
            sf_desc.read(frames=frame_duration, dtype=np.float32, always_2d=True).T,
            int(sampling_rate),
        )


def audioread_info(path: Pathlike) -> LibsndfileCompatibleAudioInfo:
    """
    Return an audio info data structure that's a compatible subset of ``pysoundfile.info()``
    that we need to create a ``Recording`` manifest.
    """
    import audioread

    # We just read the file and compute the number of samples
    # -- no other method seems fully reliable...
    with audioread.audio_open(
        str(path), backends=_available_audioread_backends()
    ) as input_file:
        shape = audioread_load(input_file)[0].shape
        if len(shape) == 1:
            num_samples = shape[0]
        else:
            num_samples = shape[1]
        return LibsndfileCompatibleAudioInfo(
            channels=input_file.channels,
            frames=num_samples,
            samplerate=int(input_file.samplerate),
            duration=num_samples / input_file.samplerate,
        )


@lru_cache(maxsize=1)
def _available_audioread_backends():
    """
    Reduces the overhead of ``audioread.audio_open()`` when called repeatedly
    by caching the results of scanning for FFMPEG etc.
    """
    import audioread

    backends = audioread.available_backends()
    logging.info(f"Using audioread. Available backends: {backends}")
    return backends


def audioread_load(
    path_or_file: Union[Pathlike, FileObject],
    offset: Seconds = 0.0,
    duration: Seconds = None,
    dtype=np.float32,
):
    """Load an audio buffer using audioread.
    This loads one block at a time, and then concatenates the results.

    This function is based on librosa:
    https://github.com/librosa/librosa/blob/main/librosa/core/audio.py#L180
    """
    import audioread

    @contextmanager
    def file_handle():
        if isinstance(path_or_file, (str, Path)):
            yield audioread.audio_open(
                path_or_file, backends=_available_audioread_backends()
            )
        else:
            yield path_or_file

    y = []
    with file_handle() as input_file:
        sr_native = input_file.samplerate
        n_channels = input_file.channels

        s_start = int(np.round(sr_native * offset)) * n_channels

        if duration is None:
            s_end = np.inf
        else:
            s_end = s_start + (int(np.round(sr_native * duration)) * n_channels)

        n = 0

        for frame in input_file:
            frame = _buf_to_float(frame, dtype=dtype)
            n_prev = n
            n = n + len(frame)

            if n < s_start:
                # offset is after the current frame
                # keep reading
                continue

            if s_end < n_prev:
                # we're off the end.  stop reading
                break

            if s_end < n:
                # the end is in this frame.  crop.
                frame = frame[: s_end - n_prev]

            if n_prev <= s_start <= n:
                # beginning is in this frame
                frame = frame[(s_start - n_prev) :]

            # tack on the current frame
            y.append(frame)

    if y:
        y = np.concatenate(y)
        if n_channels > 1:
            y = y.reshape((-1, n_channels)).T
    else:
        y = np.empty(0, dtype=dtype)

    return y, int(sr_native)


def _buf_to_float(x, n_bytes=2, dtype=np.float32):
    """Convert an integer buffer to floating point values.
    This is primarily useful when loading integer-valued wav data
    into numpy arrays.

    This function is based on librosa:
    https://github.com/librosa/librosa/blob/main/librosa/util/utils.py#L1312

    Parameters
    ----------
    x : np.ndarray [dtype=int]
        The integer-valued data buffer
    n_bytes : int [1, 2, 4]
        The number of bytes per sample in ``x``
    dtype : numeric type
        The target output type (default: 32-bit float)
    Returns
    -------
    x_float : np.ndarray [dtype=float]
        The input data buffer cast to floating point
    """

    # Invert the scale of the data
    scale = 1.0 / float(1 << ((8 * n_bytes) - 1))

    # Construct the format string
    fmt = "<i{:d}".format(n_bytes)

    # Rescale and format the data buffer
    return scale * np.frombuffer(x, fmt).astype(dtype)


def opus_info(
    path: Pathlike, force_opus_sampling_rate: Optional[int] = None
) -> LibsndfileCompatibleAudioInfo:
    samples, sampling_rate = read_opus(
        path, force_opus_sampling_rate=force_opus_sampling_rate
    )
    return LibsndfileCompatibleAudioInfo(
        channels=samples.shape[0],
        frames=samples.shape[1],
        samplerate=int(sampling_rate),
        duration=samples.shape[1] / sampling_rate,
    )


def read_opus(
    path: Pathlike,
    offset: Seconds = 0.0,
    duration: Optional[Seconds] = None,
    force_opus_sampling_rate: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    """
    Reads OPUS files either using torchaudio or ffmpeg.
    Torchaudio is faster, but if unavailable for some reason,
    we fallback to a slower ffmpeg-based implementation.

    :return: a tuple of audio samples and the sampling rate.
    """
    # TODO: Revisit using torchaudio backend for OPUS
    #       once it's more thoroughly benchmarked against ffmpeg
    #       and has a competitive I/O speed.
    #       See: https://github.com/pytorch/audio/issues/1994
    # try:
    #     return read_opus_torchaudio(
    #         path=path,
    #         offset=offset,
    #         duration=duration,
    #         force_opus_sampling_rate=force_opus_sampling_rate,
    #     )
    # except:
    return read_opus_ffmpeg(
        path=path,
        offset=offset,
        duration=duration,
        force_opus_sampling_rate=force_opus_sampling_rate,
    )


def read_opus_torchaudio(
    path: Pathlike,
    offset: Seconds = 0.0,
    duration: Optional[Seconds] = None,
    force_opus_sampling_rate: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    """
    Reads OPUS files using torchaudio.
    This is just running ``tochaudio.load()``, but we take care of extra resampling if needed.

    :return: a tuple of audio samples and the sampling rate.
    """
    audio, sampling_rate = torchaudio_load(
        path_or_fd=path, offset=offset, duration=duration
    )

    if force_opus_sampling_rate is None or force_opus_sampling_rate == sampling_rate:
        return audio, sampling_rate

    resampler = Resample(
        source_sampling_rate=sampling_rate,
        target_sampling_rate=force_opus_sampling_rate,
    )
    resampled_audio = resampler(audio)
    return resampled_audio, force_opus_sampling_rate


def read_opus_ffmpeg(
    path: Pathlike,
    offset: Seconds = 0.0,
    duration: Optional[Seconds] = None,
    force_opus_sampling_rate: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    """
    Reads OPUS files using ffmpeg in a shell subprocess.
    Unlike audioread, correctly supports offsets and durations for reading short chunks.
    Optionally, we can force ffmpeg to resample to the true sampling rate (if we know it up-front).

    :return: a tuple of audio samples and the sampling rate.
    """
    # Construct the ffmpeg command depending on the arguments passed.
    cmd = "ffmpeg -threads 1"
    sampling_rate = 48000
    # Note: we have to add offset and duration options (-ss and -t) BEFORE specifying the input
    #       (-i), otherwise ffmpeg will decode everything and trim afterwards...
    if offset > 0:
        cmd += f" -ss {offset}"
    if duration is not None:
        cmd += f" -t {duration}"
    # Add the input specifier after offset and duration.
    cmd += f" -i '{path}'"
    # Optionally resample the output.
    if force_opus_sampling_rate is not None:
        sampling_rate = force_opus_sampling_rate
    cmd += f" -ar {sampling_rate}"
    # Read audio samples directly as float32.
    cmd += " -f f32le -threads 1 pipe:1"
    # Actual audio reading.
    proc = run(cmd, shell=True, stdout=PIPE, stderr=PIPE)
    raw_audio = proc.stdout
    audio = np.frombuffer(raw_audio, dtype=np.float32)
    # Determine if the recording is mono or stereo and decode accordingly.
    try:
        channel_string = parse_channel_from_ffmpeg_output(proc.stderr)
        if channel_string == "stereo":
            new_audio = np.empty((2, audio.shape[0] // 2), dtype=np.float32)
            new_audio[0, :] = audio[::2]
            new_audio[1, :] = audio[1::2]
            audio = new_audio
        elif channel_string == "mono":
            audio = audio.reshape(1, -1)
        else:
            raise NotImplementedError(
                f"Unknown channel description from ffmpeg: {channel_string}"
            )
    except ValueError as e:
        raise AudioLoadingError(
            f"{e}\nThe ffmpeg command for which the program failed is: '{cmd}', error code: {proc.returncode}"
        )
    return audio, sampling_rate


def parse_channel_from_ffmpeg_output(ffmpeg_stderr: bytes) -> str:
    # ffmpeg will output line such as the following, amongst others:
    # "Stream #0:0: Audio: pcm_f32le, 16000 Hz, mono, flt, 512 kb/s"
    # but sometimes it can be "Stream #0:0(eng):", which we handle with regexp
    pattern = re.compile(r"^\s*Stream #0:0.*: Audio: pcm_f32le.+(mono|stereo).+\s*$")
    for line in ffmpeg_stderr.splitlines():
        try:
            line = line.decode()
        except UnicodeDecodeError:
            # Why can we get UnicodeDecoderError from ffmpeg output?
            # Because some files may contain the metadata, including a short description of the recording,
            # which may be encoded in arbitrarily encoding different than ASCII/UTF-8, such as latin-1,
            # and Python will not automatically recognize that.
            # We simply ignore these lines as they won't have any relevant information for us.
            continue
        match = pattern.match(line)
        if match is not None:
            return match.group(1)
    raise ValueError(
        f"Could not determine the number of channels for OPUS file from the following ffmpeg output "
        f"(shown as bytestring due to avoid possible encoding issues):\n{str(ffmpeg_stderr)}"
    )


def soundfile_info(path: Pathlike) -> LibsndfileCompatibleAudioInfo:
    import soundfile as sf

    if isinstance(path, Path):
        path = str(path)
    info_ = sf.info(path)
    return LibsndfileCompatibleAudioInfo(
        channels=info_.channels,
        frames=info_.frames,
        samplerate=info_.samplerate,
        duration=info_.duration,
    )


def sph_info(path: Pathlike) -> LibsndfileCompatibleAudioInfo:
    samples, sampling_rate = read_sph(path)
    return LibsndfileCompatibleAudioInfo(
        channels=samples.shape[0],
        frames=samples.shape[1],
        samplerate=int(sampling_rate),
        duration=samples.shape[1] / sampling_rate,
    )


def read_sph(
    sph_path: Pathlike, offset: Seconds = 0.0, duration: Optional[Seconds] = None
) -> Tuple[np.ndarray, int]:
    """
    Reads SPH files using sph2pipe in a shell subprocess.
    Unlike audioread, correctly supports offsets and durations for reading short chunks.

    :return: a tuple of audio samples and the sampling rate.
    """

    sph_path = Path(sph_path)

    # Construct the sph2pipe command depending on the arguments passed.
    cmd = f"sph2pipe -f wav -p -t {offset}:"

    if duration is not None:
        cmd += f"{round(offset + duration, 5)}"
    # Add the input specifier after offset and duration.
    cmd += f" {sph_path}"

    # Actual audio reading.
    try:
        proc = BytesIO(
            run(cmd, shell=True, check=True, stdout=PIPE, stderr=PIPE).stdout
        )
    except CalledProcessError as e:
        if e.returncode == 127:
            raise ValueError(
                "It seems that 'sph2pipe' binary is not installed; "
                "did you run 'lhotse install-sph2pipe'?"
            )
        else:
            raise

    import soundfile as sf

    with sf.SoundFile(proc) as sf_desc:
        audio, sampling_rate = sf_desc.read(dtype=np.float32), sf_desc.samplerate
        audio = audio.reshape(1, -1) if sf_desc.channels == 1 else audio.T

    return audio, sampling_rate


def save_audio(
    dest: Union[str, Path, BytesIO],
    src: Union[torch.Tensor, np.ndarray],
    sampling_rate: int,
    format: Optional[str] = None,
    encoding: Optional[str] = None,
) -> None:
    return get_current_audio_backend().save_audio(
        dest=dest,
        src=src,
        sampling_rate=sampling_rate,
        format=format,
        encoding=encoding,
    )


def read_audio(
    path_or_fd: Union[Pathlike, FileObject],
    offset: Seconds = 0.0,
    duration: Optional[Seconds] = None,
    force_opus_sampling_rate: Optional[int] = None,
) -> Tuple[np.ndarray, int]:
    return get_current_audio_backend().read_audio(
        path_or_fd=path_or_fd,
        offset=offset,
        duration=duration,
        force_opus_sampling_rate=force_opus_sampling_rate,
    )


def info(
    path: Union[Pathlike, BytesIO],
    force_opus_sampling_rate: Optional[int] = None,
    force_read_audio: bool = False,
) -> LibsndfileCompatibleAudioInfo:
    if force_read_audio:
        assert isinstance(
            path, (str, Path)
        ), "force_read_audio=True does not work with file-like objects"
        return AudioreadBackend().info(path_or_fd=path)
    return get_current_audio_backend().info(
        path_or_fd=path,
        force_opus_sampling_rate=force_opus_sampling_rate,
    )



"""
See https://en.data-baker.com/datasets/freeDatasets/

It is a Chinese TTS dataset, containing 12 hours of data.
"""

import logging
import re
import shutil
import tarfile
from pathlib import Path
from typing import Dict, Optional, Union

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download, safe_extract


def download_baker_zh(
    target_dir: Pathlike = ".", force_download: Optional[bool] = False
) -> Path:
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    dataset_name = "BZNSYP"
    tar_path = target_dir / f"{dataset_name}.tar.bz2"
    corpus_dir = target_dir / dataset_name
    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping {dataset_name} because {completed_detector} exists.")
        return corpus_dir
    resumable_download(
        f"https://huggingface.co/openspeech/BZNSYP/resolve/main/{dataset_name}.tar.bz2",
        filename=tar_path,
        force_download=force_download,
    )
    shutil.rmtree(corpus_dir, ignore_errors=True)
    with tarfile.open(tar_path) as tar:
        safe_extract(tar, path=target_dir)
    completed_detector.touch()

    return corpus_dir


def prepare_baker_zh(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: The RecordingSet and SupervisionSet with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    # The corpus_dir contains three sub directories
    # PhoneLabeling  ProsodyLabeling  Wave

    # Generate a mapping: utt_id -> (audio_path, audio_info, text)
    labeling_file = corpus_dir / "ProsodyLabeling" / "000001-010000.txt"
    if not labeling_file.is_file():
        raise ValueError(f"{labeling_file} does not exist")

    recordings = []
    supervisions = []
    logging.info("Started preparing. It may take 30 seconds")
    pattern = re.compile("#[12345]")
    with open(labeling_file) as f:
        try:
            while True:
                first = next(f).strip()
                pinyin = next(f).strip()
                recording_id, original_text = first.split(None, maxsplit=1)
                normalized_text = re.sub(pattern, "", original_text)
                audio_path = corpus_dir / "Wave" / f"{recording_id}.wav"

                if not audio_path.is_file():
                    logging.warning(f"No such file: {audio_path}")
                    continue
                recording = Recording.from_file(audio_path)

                segment = SupervisionSegment(
                    id=recording_id,
                    recording_id=recording_id,
                    start=0.0,
                    duration=recording.duration,
                    channel=0,
                    language="Chinese",
                    gender="female",
                    text=original_text,
                    custom={"pinyin": pinyin, "normalized_text": normalized_text},
                )
                recordings.append(recording)
                supervisions.append(segment)
        except StopIteration:
            pass

    recording_set = RecordingSet.from_recordings(recordings)
    supervision_set = SupervisionSet.from_segments(supervisions)

    recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
    validate_recordings_and_supervisions(recording_set, supervision_set)

    if output_dir is not None:
        supervision_set.to_file(output_dir / "baker_zh_supervisions_all.jsonl.gz")
        recording_set.to_file(output_dir / "baker_zh_recordings_all.jsonl.gz")

    return {"recordings": recording_set, "supervisions": supervision_set}



from abc import ABC
from dataclasses import asdict, dataclass, field
from functools import partial
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import torch

from lhotse.features.base import FeatureExtractor, register_extractor
from lhotse.utils import EPSILON, Seconds, is_module_available


@dataclass
class KaldifeatFrameOptions:
    sampling_rate: int = 16000
    frame_shift: Seconds = 0.01
    frame_length: Seconds = 0.025
    dither: float = 0.0  # default was 1.0
    preemph_coeff: float = 0.97
    remove_dc_offset: bool = True
    window_type: str = "povey"
    round_to_power_of_two: bool = True
    blackman_coeff: float = 0.42
    snip_edges: bool = False  # default was True (won't work with Lhotse)

    def to_dict(self) -> Dict[str, Any]:
        d = asdict(self)
        d["samp_freq"] = float(d.pop("sampling_rate"))
        d["frame_shift_ms"] = d.pop("frame_shift") * 1000.0
        d["frame_length_ms"] = d.pop("frame_length") * 1000.0
        return d

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "KaldifeatFrameOptions":
        data = data.copy()
        if "samp_freq" in data:
            data["sampling_rate"] = int(data.pop("samp_freq"))
        for key in ["frame_shift_ms", "frame_length_ms"]:
            if key in data:
                data[key.replace("_ms", "")] = data.pop(key) / 1000
        return KaldifeatFrameOptions(**data)


@dataclass
class KaldifeatMelOptions:
    num_bins: int = 80  # default was 23
    low_freq: float = 20.0
    high_freq: float = -400.0  # default was 0.0
    vtln_low: float = 100.0
    vtln_high: float = -500.0
    debug_mel: bool = False
    htk_mode: bool = False

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "KaldifeatMelOptions":
        return KaldifeatMelOptions(**data)


class KaldifeatExtractor(FeatureExtractor, ABC):
    """
    Base class with shared implementation for kaldifeat feature extractors.
    Derived classes are expected to set ``self.extractor`` inside ``__init__``.
    """

    def __init__(self, config: Optional[Any] = None) -> None:
        super().__init__(config=config)
        assert is_module_available(
            "kaldifeat"
        ), 'To use kaldifeat extractors, please "pip install kaldifeat" first.'

    @property
    def device(self) -> Union[str, torch.device]:
        return self.config.device

    def extract_batch(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
        lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        # kaldifeat expects a list of 1D torch tensors.
        if lengths is not None:
            samples = [x[:l] for x, l in zip(samples, lengths)]
        return self.extract(samples=samples, sampling_rate=sampling_rate)

    def extract(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        # Check for sampling rate compatibility.
        expected_sr = self.config.frame_opts.sampling_rate
        assert sampling_rate == expected_sr, (
            f"Mismatched sampling rate: extractor expects {expected_sr}, "
            f"got {sampling_rate}"
        )

        # kaldifeat expects a list of 1D torch tensors.
        # If we got a torch tensor / list of torch tensors in the input,
        # we'll also return torch tensors. If we got numpy arrays, we
        # will convert back to numpy.
        maybe_as_numpy = lambda x: x
        input_is_list = False

        if isinstance(samples, list):
            input_is_list = True
            pass  # nothing to do with `samples`
        elif samples.ndim > 1:
            samples = list(samples)
        else:
            # The user passed an array/tensor of shape (num_samples,)
            samples = [samples]

        for idx in range(len(samples)):
            if isinstance(samples[idx], np.ndarray):
                samples[idx] = torch.from_numpy(samples[idx])
                maybe_as_numpy = lambda x: x.numpy()
            if samples[idx].ndim == 2:
                # ndim could be > 1 if the input is a list of arrays/tensors.
                samples[idx] = samples[idx].squeeze()

        # Actual feature extraction.
        result = self.extractor(samples, chunk_size=self.config.chunk_size)

        # If all items are of the same shape, concatenate
        if len(result) == 1:
            if input_is_list:
                return [maybe_as_numpy(result[0])]
            else:
                return maybe_as_numpy(result[0])
        elif all(item.shape == result[0].shape for item in result[1:]):
            return maybe_as_numpy(torch.stack(result, dim=0))
        else:
            return [maybe_as_numpy(r) for r in result]

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_opts.frame_shift


@dataclass
class KaldifeatFbankConfig:
    frame_opts: KaldifeatFrameOptions = field(default_factory=KaldifeatFrameOptions)
    mel_opts: KaldifeatMelOptions = field(default_factory=KaldifeatMelOptions)
    use_energy: bool = False
    energy_floor: float = EPSILON  # default was 0.0
    raw_energy: bool = True
    htk_compat: bool = False
    use_log_fbank: bool = True
    use_power: bool = True
    device: Union[str, torch.device] = "cpu"

    # This is an extra setting compared to kaldifeat FbankOptions:
    # by default, we'll ask kaldifeat to compute the feats in chunks
    # to avoid excessive memory usage.
    chunk_size: Optional[int] = 100 * 60 * 20  # 20 minutes (assuming 100 frame/s)

    def to_dict(self) -> Dict[str, Any]:
        d = asdict(self)
        d["frame_opts"] = self.frame_opts.to_dict()
        d["mel_opts"] = self.mel_opts.to_dict()
        return d

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "KaldifeatFbankConfig":
        frame_opts = KaldifeatFrameOptions.from_dict(data.pop("frame_opts"))
        mel_opts = KaldifeatMelOptions.from_dict(data.pop("mel_opts"))
        return KaldifeatFbankConfig(frame_opts=frame_opts, mel_opts=mel_opts, **data)


@register_extractor
class KaldifeatFbank(KaldifeatExtractor):
    """Log Mel energy filter bank feature extractor based on ``kaldifeat`` package."""

    name = "kaldifeat-fbank"
    config_type = KaldifeatFbankConfig

    def __init__(self, config: Optional[KaldifeatFbankConfig] = None) -> None:
        super().__init__(config)
        import kaldifeat

        self.extractor = kaldifeat.Fbank(
            kaldifeat.FbankOptions.from_dict(self.config.to_dict())
        )

    def feature_dim(self, sampling_rate: int) -> int:
        return self.config.mel_opts.num_bins

    @staticmethod
    def mix(
        features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float
    ) -> np.ndarray:
        return np.log(
            np.maximum(
                # protection against log(0); max with EPSILON is adequate since these are energies (always >= 0)
                EPSILON,
                np.exp(features_a) + energy_scaling_factor_b * np.exp(features_b),
            )
        )

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        return float(np.sum(np.exp(features)))


@dataclass
class KaldifeatMfccConfig:
    frame_opts: KaldifeatFrameOptions = field(default_factory=KaldifeatFrameOptions)
    mel_opts: KaldifeatMelOptions = field(
        default_factory=partial(KaldifeatMelOptions, num_bins=23)
    )
    num_ceps: int = 13
    use_energy: bool = False
    energy_floor: float = EPSILON  # default was 0.0
    raw_energy: bool = True
    cepstral_lifter: float = 22.0
    htk_compat: bool = False
    device: Union[str, torch.device] = "cpu"

    # This is an extra setting compared to kaldifeat FbankOptions:
    # by default, we'll ask kaldifeat to compute the feats in chunks
    # to avoid excessive memory usage.
    chunk_size: Optional[int] = 1000

    def to_dict(self) -> Dict[str, Any]:
        d = asdict(self)
        d["frame_opts"] = self.frame_opts.to_dict()
        d["mel_opts"] = self.mel_opts.to_dict()
        return d

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "KaldifeatMfccConfig":
        frame_opts = KaldifeatFrameOptions.from_dict(data.pop("frame_opts"))
        mel_opts = KaldifeatMelOptions.from_dict(data.pop("mel_opts"))
        return KaldifeatMfccConfig(frame_opts=frame_opts, mel_opts=mel_opts, **data)


@register_extractor
class KaldifeatMfcc(KaldifeatExtractor):
    """MFCC feature extractor based on ``kaldifeat`` package."""

    name = "kaldifeat-mfcc"
    config_type = KaldifeatMfccConfig

    def __init__(self, config: Optional[KaldifeatMfccConfig] = None) -> None:
        super().__init__(config)
        import kaldifeat

        self.extractor = kaldifeat.Mfcc(
            kaldifeat.MfccOptions.from_dict(self.config.to_dict())
        )

    def feature_dim(self, sampling_rate: int) -> int:
        return self.config.num_ceps



from typing import Dict, List, Optional, Tuple, Union

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.icmcasr import prepare_icmcasr
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
@click.option(
    "--mic",
    type=click.Choice(["ihm", "sdm", "mdm"]),
    default="ihm",
    help="Microphone type.",
)
def icmcasr(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    mic: str = "ihm",
    num_jobs: int = 1,
):
    """ICMC-ASR data preparation."""
    prepare_icmcasr(
        corpus_dir=corpus_dir,
        output_dir=output_dir,
        mic=mic,
        num_jobs=num_jobs,
    )
